"""Render simple text and formatted documents efficiently.

Three layout classes are provided:

:py:class:`~pyglet.text.layout.TextLayout`
    The entire document is laid out before it is rendered.  The layout will
    be grouped with other layouts in the same batch (allowing for efficient
    rendering of multiple layouts).

    Any change to the layout or document,
    and even querying some properties, will cause the entire document
    to be laid out again.

:py:class:`~pyglet.text.layout.ScrollableTextLayout`
    Based on :py:func:`~pyglet.text.layout.TextLayout`.

    A separate group is used for layout which crops the contents of the
    layout to the layout rectangle.  Additionally, the contents of the
    layout can be "scrolled" within that rectangle with the ``view_x`` and
    ``view_y`` properties.

:py:class:`~pyglet.text.layout.IncrementalTextLayout`
    Based on :py:class:`~pyglet.text.layout.ScrollableTextLayout`.

    When the layout or document are modified, only the affected regions
    are laid out again.  This permits efficient interactive editing and
    styling of text.

    Only the visible portion of the layout is actually rendered; as the
    viewport is scrolled additional sections are rendered and discarded as
    required.  This permits efficient viewing and editing of large documents.

    Additionally, this class provides methods for locating the position of a
    caret in the document, and for displaying interactive text selections.

All three layout classes can be used with either :py:class:`~pyglet.text.document.UnformattedDocument` or
:py:class:`~pyglet.text.document.FormattedDocument`, and can be either single-line or ``multiline``.  The
combinations of these options effectively provides 12 different text display
possibilities.

Style attributes
================

The following character style attribute names are recognised by the layout
classes.  Data types and units are as specified.

Where an attribute is marked "as a distance" the value is assumed to be
in pixels if given as an int or float, otherwise a string of the form
``"0u"`` is required, where ``0`` is the distance and ``u`` is the unit; one
of ``"px"`` (pixels), ``"pt"`` (points), ``"pc"`` (picas), ``"cm"``
(centimeters), ``"mm"`` (millimeters) or ``"in"`` (inches).  For example,
``"14pt"`` is the distance covering 14 points, which at the default DPI of 96
is 18 pixels.

``font_name``
    Font family name, as given to :py:func:`pyglet.font.load`.
``font_size``
    Font size, in points.
``weight``
    String.
``italic``
    Boolean.
``underline``
    4-tuple of ints in range (0, 255) giving RGBA underline color, or None
    (default) for no underline.
``kerning``
    Additional space to insert between glyphs, as a distance.  Defaults to 0.
``baseline``
    Offset of glyph baseline from line baseline, as a distance.  Positive
    values give a superscript, negative values give a subscript.  Defaults to
    0.
``color``
    4-tuple of ints in range (0, 255) giving RGBA text color
``background_color``
    4-tuple of ints in range (0, 255) giving RGBA text background color; or
    ``None`` for no background fill.

The following paragraph style attribute names are recognised.  Note
that paragraph styles are handled no differently from character styles by the
document: it is the application's responsibility to set the style over an
entire paragraph, otherwise results are undefined.

``align``
    ``left`` (default), ``center`` or ``right``.
``indent``
    Additional horizontal space to insert before the first glyph of the
    first line of a paragraph, as a distance.
``leading``
    Additional space to insert between consecutive lines within a paragraph,
    as a distance.  Defaults to 0.
``line_spacing``
    Distance between consecutive baselines in a paragraph, as a distance.
    Defaults to ``None``, which automatically calculates the tightest line
    spacing for each line based on the font ascent and descent.
``margin_left``
    Left paragraph margin, as a distance.
``margin_right``
    Right paragraph margin, as a distance.
``margin_top``
    Margin above paragraph, as a distance.
``margin_bottom``
    Margin below paragraph, as a distance.  Adjacent margins do not collapse.
``tab_stops``
    List of horizontal tab stops, as distances, measured from the left edge of
    the text layout.  Defaults to the empty list.  When the tab stops
    are exhausted, they implicitly continue at 50 pixel intervals.
``wrap``
    ``char``, ``word``, True (default) or False.  The boundaries at which to
    wrap text to prevent it overflowing a line.  With ``char``, the line
    wraps anywhere in the text; with ``word`` or True, the line wraps at
    appropriate boundaries between words; with False the line does not wrap,
    and may overflow the layout width.

Other attributes can be used to store additional style information within the
document; they will be ignored by the built-in text classes.

"""
from __future__ import annotations

from pyglet.text.layout.base import (
    TextDecorationGroup,
    TextLayout,
    TextLayoutGroup,
    decoration_fragment_source,
    decoration_vertex_source,
    get_default_decoration_shader,
    get_default_image_layout_shader,
    get_default_layout_shader,
    layout_fragment_image_source,
    layout_fragment_source,
    layout_vertex_source,
)
from pyglet.text.layout.incremental import (
    IncrementalTextDecorationGroup,
    IncrementalTextLayout,
    IncrementalTextLayoutGroup,
)
from pyglet.text.layout.scrolling import ScrollableTextDecorationGroup, ScrollableTextLayout, ScrollableTextLayoutGroup

__all__ = ["TextLayout", "IncrementalTextLayout", "ScrollableTextLayout", "TextLayoutGroup", "TextDecorationGroup",
           "IncrementalTextLayoutGroup", "IncrementalTextDecorationGroup", "ScrollableTextLayoutGroup",
           "ScrollableTextDecorationGroup", "get_default_layout_shader", "get_default_image_layout_shader",
           "get_default_decoration_shader", "decoration_fragment_source", "layout_fragment_image_source",
           "layout_fragment_source", "layout_vertex_source", "decoration_vertex_source"]



from __future__ import annotations

import warnings
from ctypes import Structure, c_ushort, c_ulong, c_int, c_void_p, POINTER
import pyglet.lib
from pyglet.libs.x11.xlib import Visual

lib = None
try:
    lib = pyglet.lib.load_library("Xrender")
except ImportError:
    if pyglet.options.debug_lib:
        warnings.warn("XRender could not be loaded.")


class XRenderDirectFormat(Structure):
    _fields_ = [
        ('red', c_ushort),
        ('redMask', c_ushort),
        ('green', c_ushort),
        ('greenMask', c_ushort),
        ('blue', c_ushort),
        ('blueMask', c_ushort),
        ('alpha', c_ushort),
        ('alphaMask', c_ushort),
    ]

class XRenderPictFormat(Structure):
    _fields_ = [
        ('id', c_ulong),
        ('type', c_int),
        ('depth', c_int),
        ('direct', XRenderDirectFormat),
        ('colormap', c_ulong),
    ]

# XRenderFindVisualFormat(Display *dpy, Visual *visual)
try:
    XRenderFindVisualFormat = lib.XRenderFindVisualFormat
    XRenderFindVisualFormat.argtypes = [c_void_p, POINTER(Visual)]
    XRenderFindVisualFormat.restype = POINTER(XRenderPictFormat)
except ValueError:
    XRenderFindVisualFormat = None



"""This example shows how to use multiple textures in a shader with Pyglet."""
from __future__ import annotations

from typing import TYPE_CHECKING

import pyglet
from pyglet.gl import glActiveTexture, GL_TEXTURE0, glBindTexture, glEnable, GL_BLEND, glBlendFunc, glDisable, \
    glClearColor, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA

if TYPE_CHECKING:
    from pyglet.image import Texture, AbstractImage
    from pyglet.graphics import Batch, Group
    from pyglet.graphics.shader import ShaderProgram

vertex_source = """#version 150 core
    in vec3 translate;
    in vec4 colors;
    in vec3 tex_coords;
    in vec2 scale;
    in vec3 position;
    in float rotation;

    out vec4 vertex_colors;
    out vec3 texture_coords;

    uniform WindowBlock
    {
        mat4 projection;
        mat4 view;
    } window;

    mat4 m_scale = mat4(1.0);
    mat4 m_rotation = mat4(1.0);
    mat4 m_translate = mat4(1.0);

    void main()
    {
        m_scale[0][0] = scale.x;
        m_scale[1][1] = scale.y;
        m_translate[3][0] = translate.x;
        m_translate[3][1] = translate.y;
        m_translate[3][2] = translate.z;
        m_rotation[0][0] =  cos(-radians(rotation)); 
        m_rotation[0][1] =  sin(-radians(rotation));
        m_rotation[1][0] = -sin(-radians(rotation));
        m_rotation[1][1] =  cos(-radians(rotation));

        gl_Position = window.projection * window.view * m_translate * m_rotation * m_scale * vec4(position, 1.0);

        vertex_colors = colors;
        texture_coords = tex_coords;
    }
"""

fragment_source = """#version 150 core
    in vec4 vertex_colors;
    in vec3 texture_coords;
    out vec4 final_colors;

    uniform sampler2D kitten_texture;
    uniform sampler2D pyglet_texture;

    void main()
    {
        final_colors = texture(kitten_texture, texture_coords.xy) * texture(pyglet_texture, texture_coords.xy) * 
        vertex_colors;
    }
"""


class MultiTextureSpriteGroup(pyglet.sprite.SpriteGroup):
    """A sprite group that uses multiple active textures."""

    def __init__(self, textures: dict[str, Texture], blend_src: int, blend_dest: int,
                 program: ShaderProgram | None = None, parent: Group | None = None) -> None:
        """Create a sprite group for multiple textures and samplers.

        All textures must share the same target type.

        Args:
            textures:
                A dictionary of textures, with the keys being the GLSL sampler name.
            blend_src:
                OpenGL blend source mode; for example, ``GL_SRC_ALPHA``.
            blend_dest:
                OpenGL blend destination mode; for example, ``GL_ONE_MINUS_SRC_ALPHA``.
            parent:
                Optional parent group.
        """
        self.textures = textures
        texture = list(self.textures.values())[0]
        self.target = texture.target
        super().__init__(texture, blend_src, blend_dest, program, parent)

    def set_state(self) -> None:
        self.program.use()

        for idx, name in enumerate(self.textures):
            self.program[name] = idx

        for i, texture in enumerate(self.textures.values()):
            glActiveTexture(GL_TEXTURE0 + i)
            glBindTexture(self.target, texture.id)

        glEnable(GL_BLEND)
        glBlendFunc(self.blend_src, self.blend_dest)

    def unset_state(self) -> None:
        glDisable(GL_BLEND)
        self.program.stop()
        glActiveTexture(GL_TEXTURE0)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.texture}-{self.texture.id})'

    def __eq__(self, other: object | Group | MultiTextureSpriteGroup) -> bool:
        return (other.__class__ is self.__class__ and
                self.program is other.program and
                self.textures == other.textures and
                self.blend_src == other.blend_src and
                self.blend_dest == other.blend_dest)

    def __hash__(self) -> int:
        return hash((id(self.parent),
                     id(self.textures),
                     self.blend_src, self.blend_dest))


class MultiTextureSprite(pyglet.sprite.Sprite):
    """An example of a Sprite that can utilize multiple textures for an effect."""
    group_class = MultiTextureSpriteGroup

    def __init__(self,
                 images: dict[str, AbstractImage],
                 x: float = 0, y: float = 0, z: float = 0,
                 blend_src: int = GL_SRC_ALPHA,
                 blend_dest: int = GL_ONE_MINUS_SRC_ALPHA,
                 batch: Batch | None = None,
                 group: Group | None = None,
                 subpixel: bool = False,
                 program: ShaderProgram | None = None) -> None:
        """Create a Sprite instance.

        Args:
            images:
                A dictionary of images, with the keys being the GLSL sampler name.
            x:
                X coordinate of the sprite.
            y:
                Y coordinate of the sprite.
            z:
                Z coordinate of the sprite.
            blend_src:
                OpenGL blend source mode.  The default is suitable for
                compositing sprites drawn from back-to-front.
            blend_dest:
                OpenGL blend destination mode.  The default is suitable for
                compositing sprites drawn from back-to-front.
            batch:
                Optional batch to add the sprite to.
            group:
                Optional parent group of the sprite.
            subpixel:
                Allow floating-point coordinates for the sprite. By default,
                coordinates are restricted to integer values.
            program:
                A specific shader program to initialize the sprite with. By default, a pre-made shader will be chosen
                based on the texture type passed.
        """
        # Ensure the images are textures.
        self.textures: dict[str, Texture] = {name: img.get_texture() for name, img in images.items()}
        assert all(tex.target for tex in self.textures.values()) is True, "All textures need to be the same target."

        # Use first image as base.
        super().__init__(list(self.textures.values())[0], x, y, z, blend_src, blend_dest, batch, group, subpixel,
                         program)

    def get_sprite_group(self) -> MultiTextureSpriteGroup:
        return self.group_class(self.textures, self._blend_src, self._blend_dest, self._program, self._user_group)


window = pyglet.window.Window()

# Set example resource path.
pyglet.resource.path = ['../resources']
pyglet.resource.reindex()

# Load example image from resource path.
# Resource will load images into a texture atlas, disable atlas loading, so they can be separate textures.
pyglet_image = pyglet.resource.image("pyglet.png", atlas=False)
kitten_image = pyglet.resource.image("kitten.jpg", atlas=False)

window.set_size(kitten_image.width, kitten_image.height)

# Batching allows rendering groups of objects all at once instead of drawing one by one.
batch = pyglet.graphics.Batch()

# Create our new shader that handles both texture sprites.
multi_vert_shader = pyglet.graphics.shader.Shader(vertex_source, 'vertex')
multi_frag_shader = pyglet.graphics.shader.Shader(fragment_source, 'fragment')

# Our new shader just multiplies both images together.
multitex_shader_program = pyglet.graphics.shader.ShaderProgram(multi_vert_shader, multi_frag_shader)

# Give our shader names and textures.
shader_images = {
    "kitten_texture": kitten_image,
    "pyglet_texture": pyglet_image,
}

sprite = MultiTextureSprite(shader_images,
                            x=0,
                            y=0,
                            batch=batch,
                            program=multitex_shader_program)

glClearColor(1.0, 1.0, 1.0, 1.0)


@window.event
def on_draw():
    window.clear()
    batch.draw()


pyglet.app.run()



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from pyglet import graphics
from pyglet.gl import (
    GL_BLEND,
    GL_ONE_MINUS_SRC_ALPHA,
    GL_SRC_ALPHA,
    GL_TEXTURE0,
    glActiveTexture,
    glBindTexture,
    glBlendFunc,
    glDisable,
    glEnable,
)
from pyglet.text.layout.base import TextLayout

if TYPE_CHECKING:
    from pyglet.customtypes import AnchorX, AnchorY
    from pyglet.graphics import Batch
    from pyglet.graphics.shader import ShaderProgram
    from pyglet.image import Texture
    from pyglet.text.document import AbstractDocument


class ScrollableTextLayoutGroup(graphics.Group):
    """Default rendering group for :py:class:`~pyglet.text.layout.ScrollableTextLayout`.

    The group maintains internal state for specifying the viewable
    area, and for scrolling. Because the group has internal state
    specific to the text layout, the group is never shared.
    """
    scissor_area: ClassVar[tuple[int, int, int, int]] = 0, 0, 0, 0

    def __init__(self, texture: Texture, program: ShaderProgram, order: int = 1,  # noqa: D107
                 parent: graphics.Group | None = None) -> None:

        super().__init__(order=order, parent=parent)
        self.texture = texture
        self.program = program

    def set_state(self) -> None:
        self.program.use()
        self.program["scissor"] = True
        self.program["scissor_area"] = self.scissor_area

        glActiveTexture(GL_TEXTURE0)
        glBindTexture(self.texture.target, self.texture.id)

        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    def unset_state(self) -> None:
        glDisable(GL_BLEND)
        self.program.stop()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.texture})"

    def __eq__(self, other: object) -> bool:
        return self is other

    def __hash__(self) -> int:
        return id(self)


class ScrollableTextDecorationGroup(graphics.Group):
    """Create a text decoration rendering group.

    The group is created internally when a :py:class:`~pyglet.text.Label`
    is created; applications usually do not need to explicitly create it.
    """

    scissor_area: ClassVar[tuple[int, int, int, int]] = 0, 0, 0, 0

    def __init__(self, program: ShaderProgram, order: int = 0, parent: graphics.Group | None = None) -> None:  # noqa: D107
        super().__init__(order=order, parent=parent)
        self.program = program

    def set_state(self) -> None:
        self.program.use()
        self.program["scissor"] = True
        self.program["scissor_area"] = self.scissor_area

        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    def unset_state(self) -> None:
        glDisable(GL_BLEND)
        self.program.stop()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(scissor={self.scissor_area})"

    def __eq__(self, other: object) -> bool:
        return self is other

    def __hash__(self) -> int:
        return id(self)


class ScrollableTextLayout(TextLayout):
    """Display text in a scrollable viewport.

    This class does not display a scrollbar or handle scroll events; it merely
    clips the text that would be drawn in :py:func:`~pyglet.text.layout.TextLayout`
    to the bounds of the layout given by ``x``, ``y``, ``width`` and ``height``;
    and offsets the text by a scroll offset.

    Use ``view_x`` and ``view_y`` to scroll the text within the viewport.

    Attributes:
        group_class:
            Default group used to set the state for all glyphs.
        decoration_class:
            Default group used to set the state for all decorations including background colors and underlines.
    """

    group_class: ClassVar[type[ScrollableTextLayoutGroup]] = ScrollableTextLayoutGroup
    decoration_class: ClassVar[type[ScrollableTextDecorationGroup]] = ScrollableTextDecorationGroup

    _translate_x: int = 0
    _translate_y: int = 0

    def __init__(self, document: AbstractDocument,  # noqa: D107
                 x: float = 0, y: float = 0, z: float = 0,
                 width: int = None, height: int = None,
                 anchor_x: AnchorX = 'left', anchor_y: AnchorY = 'bottom', rotation: float = 0, multiline: bool = False,
                 dpi: float | None = None, batch: Batch | None = None, group: graphics.Group | None = None,
                 program: ShaderProgram | None = None, wrap_lines: bool = True) -> None:

        if width is None or height is None:
            msg = "Invalid size. ScrollableTextLayout width or height cannot be None."
            raise Exception(msg)

        super().__init__(document, x, y, z, width, height, anchor_x, anchor_y, rotation, multiline, dpi, batch, group,
                         program, wrap_lines)

    def _update_scissor_area(self) -> None:
        if not self.document.text:
            return

        area = (self.left, self.bottom, self._width, self._height)

        for group in self.group_cache.values():
            group.scissor_area = area

        self.background_decoration_group.scissor_area = area
        self.foreground_decoration_group.scissor_area = area

    def _update(self) -> None:
        super()._update()
        self._update_scissor_area()

    # Properties

    @property
    def x(self) -> float:
        return self._x

    @x.setter
    def x(self, x: float) -> None:
        super()._set_x(x)
        self._update_scissor_area()

    @property
    def y(self) -> float:
        return self._y

    @y.setter
    def y(self, y: float) -> None:
        super()._set_y(y)
        self._update_scissor_area()

    @property
    def z(self) -> float:
        return self._z

    @z.setter
    def z(self, z: float) -> None:
        super()._set_z(z)
        self._update_scissor_area()

    @property
    def position(self) -> tuple[float, float, float]:
        return self._x, self._y, self._z

    @position.setter
    def position(self, position: tuple[float, float, float]) -> None:
        super()._set_position(position)
        self._update_scissor_area()

    @property
    def anchor_x(self) -> AnchorX:
        return self._anchor_x

    @anchor_x.setter
    def anchor_x(self, anchor_x: AnchorX) -> None:
        self._anchor_x = anchor_x
        self._update_anchor()
        self._update_scissor_area()
        self._update_view_translation()

    @property
    def anchor_y(self) -> AnchorY:
        return self._anchor_y

    @anchor_y.setter
    def anchor_y(self, anchor_y: AnchorY) -> None:
        self._anchor_y = anchor_y
        self._update_anchor()
        self._update_scissor_area()
        self._update_view_translation()

    def _get_bottom_anchor(self) -> float:
        """Returns the anchor for the Y axis from the bottom."""
        height = self._height
        if self._content_valign == "top":
            offset = min(0, self._height)
        elif self._content_valign == "bottom":
            offset = 0
        elif self._content_valign == "center":
            offset = min(0, self._height) // 2
        else:
            msg = '`content_valign` must be either "top", "bottom", or "center".'
            raise Exception(msg)

        if self._anchor_y == "top":
            return -height + offset
        if self._anchor_y == "baseline":
            return -height + self._ascent
        if self._anchor_y == "bottom":
            return 0
        if self._anchor_y == "center":
            if self._line_count == 1 and self._height is None:
                # This "looks" more centered than considering all of the descent.
                return (self._ascent // 2 - self._descent // 4) - height

            return offset - height // 2

        msg = '`anchor_y` must be either "top", "bottom", "center", or "baseline".'
        raise Exception(msg)

    def _update_view_translation(self) -> None:
        # Offset of content within viewport
        for _vertex_list in self._vertex_lists:
            _vertex_list.view_translation[:] = (-self._translate_x, -self._translate_y, 0) * _vertex_list.count

    @property
    def view_x(self) -> int:
        """Horizontal scroll offset.

        The initial value is 0, and the left edge of the text will touch the left
        side of the layout bounds.  A positive value causes the text to "scroll"
        to the right.  Values are automatically clipped into the range
        ``[0, content_width - width]``
        """
        return self._translate_x

    @view_x.setter
    def view_x(self, view_x: int) -> None:
        translation = max(0, min(self._content_width - self._width, view_x))
        if translation != self._translate_x:
            self._translate_x = translation
            self._update_view_translation()

    @property
    def view_y(self) -> int:
        """Vertical scroll offset.

        The initial value is 0, and the top of the text will touch the top of the
        layout bounds (unless the content height is less than the layout height,
        in which case `content_valign` is used).

        A negative value causes the text to "scroll" upwards.  Values outside of
        the range ``[height - content_height, 0]`` are automatically clipped in
        range.
        """
        return self._translate_y

    @view_y.setter
    def view_y(self, view_y: int) -> None:
        # view_y must be negative.
        translation = min(0, max(self.height - self._content_height, view_y))
        if translation != self._translate_y:
            self._translate_y = translation
            self._update_view_translation()



"""
Test internals of the DirectSound media driver.
"""
import ctypes
import math
import random
import time

import pytest
from ...annotations import skip_if_continuous_integration, require_platform, Platform

try:
    from pyglet.media.codecs import AudioFormat
    from pyglet.media.drivers.directsound.adaptation import _db2gain, _gain2db
    from pyglet.media.drivers.directsound.interface import DirectSoundDriver
except ImportError:
    AudioFormat = DirectSoundDriver = _db2gain = _gain2db = None

pytestmark = [skip_if_continuous_integration(), require_platform(Platform.WINDOWS)]


def almost_equal(a, b, e=0.0001):
    assert abs(a - b) <= e
    return True


def iter_almost_equal(a, b, e=0.0001):
    for x, y in zip(a, b):
        assert abs(x - y) <= e
    return True


def random_normalized_vector():
    vector = [random.uniform(-1.0, 1.0) for _ in range(3)]

    length = math.sqrt(sum(x ** 2 for x in vector))
    return [x / length for x in vector]


@pytest.fixture
def driver():
    return DirectSoundDriver()


@pytest.fixture(params=[(1, 8, 22050),
                        (2, 8, 22050),
                        (1, 16, 22050),
                        (2, 16, 22050),
                        (1, 16, 44100),
                        (2, 16, 44100)])
def audio_format(request):
    return AudioFormat(*request.param)


@pytest.fixture(params=[(1, 8, 22050),
                        (1, 16, 22050),
                        (1, 16, 44100)])
def audio_format_3d(request):
    return AudioFormat(*request.param)


@pytest.fixture
def buffer_(driver, audio_format):
    buffer = driver.create_buffer(audio_format, int(audio_format.bytes_per_second * 1.0))
    yield buffer
    buffer.delete()


@pytest.fixture
def buffer_3d(driver, audio_format_3d):
    buffer = driver.create_buffer(audio_format_3d, int(audio_format_3d.bytes_per_second * 1.0))
    assert buffer.is3d
    yield buffer
    buffer.delete()


@pytest.fixture
def filled_buffer(audio_format, buffer_):
    pointer = buffer_.lock(0, 1024)
    if audio_format.sample_size == 8:
        c = 0x80
    else:
        c = 0x00
    ctypes.memset(pointer.audio_ptr_1, c, pointer.audio_length_1.value)
    if pointer.audio_length_2.value > 0:
        ctypes.memset(pointer.audio_ptr_2, c, pointer.audio_length_2.value)

    return buffer_


@pytest.fixture
def listener(driver):
    listener = driver.create_listener()
    yield listener
    listener.delete()


def test_driver_create():
    driver = DirectSoundDriver()
    del driver


def test_create_buffer(driver, audio_format):
    buf = driver.create_buffer(audio_format, int(audio_format.bytes_per_second * 1.0))
    del buf


def test_buffer_volume(buffer_):
    vol = 0.0
    while vol <= 1.0:
        listener.volume = _gain2db(vol)
        assert almost_equal(listener.volume, _gain2db(vol), 0.05)
        vol += 0.05


def test_buffer_current_position_empty_buffer(buffer_):
    buffer_.current_position = 0
    assert buffer_.current_position == (0, 0)


def test_buffer_current_position_filled_buffer(filled_buffer):
    filled_buffer.current_position = 512
    assert filled_buffer.current_position == (512, 512)


def test_buffer_is3d(audio_format, buffer_):
    assert buffer_.is3d == (audio_format.channels == 1)


def test_buffer_position(buffer_):
    for _ in range(10):
        position = [random.uniform(-10.0, 10.0) for _ in range(3)]
        buffer_.position = position
        if buffer_.is3d:
            assert iter_almost_equal(buffer_.position, position)
        else:
            assert buffer_.position == (0, 0, 0)


def test_buffer_min_distance(buffer_):
    for _ in range(10):
        distance = random.uniform(0.0, 100.0)
        buffer_.min_distance = distance
        if buffer_.is3d:
            assert almost_equal(buffer_.min_distance, distance)
        else:
            assert buffer_.min_distance == 0


def test_buffer_max_distance(buffer_):
    for _ in range(10):
        distance = random.uniform(0.0, 100.0)
        buffer_.max_distance = distance
        if buffer_.is3d:
            assert almost_equal(buffer_.max_distance, distance)
        else:
            assert buffer_.max_distance == 0


def test_buffer_cone_orienation(buffer_):
    for _ in range(10):
        orientation = random_normalized_vector()
        buffer_.cone_orientation = orientation
        if buffer_.is3d:
            assert iter_almost_equal(buffer_.cone_orientation, orientation)
        else:
            assert buffer_.cone_orientation == (0, 0, 0)


def test_buffer_cone_angles(buffer_):
    for _ in range(10):
        angle1 = random.randint(0, 360)
        angle2 = random.randint(0, 360)
        inside = min(angle1, angle2)
        outside = max(angle1, angle2)
        buffer_.set_cone_angles(inside, outside)
        result = buffer_.cone_angles
        if buffer_.is3d:
            assert result.inside == inside
            assert result.outside == outside
        else:
            assert result.inside == 0
            assert result.outside == 0


def test_buffer_cone_outside_volume(buffer_):
    for _ in range(10):
        volume = _gain2db(random.uniform(0.0, 1.0))
        buffer_.cone_outside_volume = volume
        if buffer_.is3d:
            assert almost_equal(buffer_.cone_outside_volume, volume)
        else:
            assert buffer_.cone_outside_volume == 0


def test_buffer_frequency(buffer_):
    for _ in range(10):
        freq = random.randint(100, 100000)
        buffer_.frequency = freq
        assert buffer_.frequency == freq


def test_buffer_lock_unlock(buffer_):
    size = 1024
    pointer = buffer_.lock(0, size)
    assert pointer.audio_length_1.value + pointer.audio_length_2.value == size
    buffer_.unlock(pointer)


def test_buffer_play_stop(filled_buffer):
    assert filled_buffer.current_position[0] == 0
    filled_buffer.play()
    for _ in range(100):
        assert filled_buffer.is_playing
        if filled_buffer.current_position[0] > 0:
            break
        else:
            time.sleep(0.001)
    else:
        pytest.fail("Did not advance position in buffer while playing.")

    filled_buffer.stop()
    assert not filled_buffer.is_playing
    pos = filled_buffer.current_position
    for _ in range(10):
        assert filled_buffer.current_position == pos
        time.sleep(0.001)


def test_create_listener(driver):
    listener = driver.create_listener()
    del listener


def test_listener_position(listener):
    for _ in range(10):
        position = [random.uniform(-10.0, 10.0) for _ in range(3)]
        listener.position = position
        assert iter_almost_equal(listener.position, position)


def test_listener_orientation(listener):
    for _ in range(10):
        front = random_normalized_vector()
        top = random_normalized_vector()
        orientation = front + top

        listener.orientation = orientation
        # Only testing first 3, as random values might be adjusted by DS to be correct angles
        assert iter_almost_equal(listener.orientation[:3], orientation[:3])



# Test the Objective-C integration.
import unittest
import weakref
import gc

import pyglet
from tests.annotations import require_platform, Platform

pytestmark = require_platform(Platform.OSX)

if pyglet.compat_platform in ("darwin",):
    from pyglet.libs.darwin import AutoReleasePool, ObjCSubclass, ObjCClass
    from pyglet.libs.darwin.cocoapy.runtime import get_cached_instances, send_super

    NSObject = ObjCClass('NSObject')


class ObjCIntegrationTest(unittest.TestCase):
    def test_objc_leak_gc(self):
        """Test deleting """
        start_count = len(get_cached_instances())
        test_object = NSObject.alloc().init()
        del test_object

        gc.collect()

        self.assertEqual(len(get_cached_instances()), start_count)

    def test_objc_leak_cache_release_manual_delete(self):
        start_count = len(get_cached_instances())
        test_object = NSObject.alloc().init()

        self.assertIs(test_object._retained, True)

        test_object.release()

        self.assertIs(test_object._retained, False)

        del test_object

        self.assertEqual(len(get_cached_instances()), start_count)

    def test_objc_subclass_dealloc_release(self):
        start_count = len(get_cached_instances())

        allocated = True

        class MyCustomObjectTestRelease_Implementation:
            MyCustomObjectTestRelease = ObjCSubclass("NSObject", "MyCustomObjectTestRelease")

            @MyCustomObjectTestRelease.method('v')
            def dealloc(self) -> None:
                nonlocal allocated
                allocated = False
                send_super(self, 'dealloc')

        MyCustomObjectTestRelease = ObjCClass('MyCustomObjectTestRelease')

        instance = MyCustomObjectTestRelease.alloc().init()
        instance.release()

        self.assertEqual(allocated, False)

        del instance

        self.assertEqual(len(get_cached_instances()), start_count)

    def test_objc_subclass_dealloc_autorelease(self):
        """Pytest doesn't like reusing an ObjCClass in the setUpClass. Just make a new one for testing."""
        start_count = len(get_cached_instances())

        allocated = True

        class MyCustomObjectTestAutorelease_Implementation:
            MyCustomObjectTestAutorelease = ObjCSubclass("NSObject", "MyCustomObjectTestAutorelease")

            @MyCustomObjectTestAutorelease.method('v')
            def dealloc(self) -> None:
                nonlocal allocated
                allocated = False
                send_super(self, 'dealloc')

        MyCustomObjectTestAutorelease = ObjCClass('MyCustomObjectTestAutorelease')

        with AutoReleasePool():
            MyCustomObjectTestAutorelease.alloc().init().autorelease()

        self.assertEqual(allocated, False)
        self.assertEqual(len(get_cached_instances()), start_count)

    def test_objc_association(self):
        """Make sure associated Python objects don't get GC'd until the ObjC instance does."""
        with AutoReleasePool():
            test_object = NSObject.alloc().init()

            class Data:
                pass

            data = Data()

            data_ref = weakref.ref(data)

            test_object.associate("data", data)

            del data

            gc.collect()

            self.assertIs(test_object.data, data_ref())

            del test_object
            gc.collect()

        self.assertIsNone(data_ref())



from ctypes import (
    POINTER,
    Structure,
    c_buffer,
    c_byte,
    c_float,
    c_long,
    c_short,
    c_uint32,
    c_ulong,
    c_void_p,
    c_wchar,
    c_wchar_p,
    cast,
    create_string_buffer,
    memmove,
    windll,
)
from ctypes.wintypes import BOOL, BYTE, INT, UINT, ULONG

from pyglet.image import Animation, AnimationFrame, ImageData, byref, c_int, c_uint, sizeof
from pyglet.image.codecs import ImageDecodeException, ImageDecoder
from pyglet.libs.win32 import _kernel32 as kernel32
from pyglet.libs.win32 import _ole32 as ole32
from pyglet.libs.win32.com import pIUnknown
from pyglet.libs.win32.constants import GMEM_MOVEABLE
from pyglet.libs.win32.types import LONG_PTR

gdiplus = windll.gdiplus

REAL = c_float

PixelFormat1bppIndexed    = 196865
PixelFormat4bppIndexed    = 197634
PixelFormat8bppIndexed    = 198659
PixelFormat16bppGrayScale = 1052676
PixelFormat16bppRGB555    = 135173
PixelFormat16bppRGB565    = 135174
PixelFormat16bppARGB1555  = 397319
PixelFormat24bppRGB       = 137224
PixelFormat32bppRGB       = 139273
PixelFormat32bppARGB      = 2498570
PixelFormat32bppPARGB     = 925707
PixelFormat48bppRGB       = 1060876
PixelFormat64bppARGB      = 3424269
PixelFormat64bppPARGB     = 29622286
PixelFormatMax            = 15

ImageLockModeRead = 1
ImageLockModeWrite = 2
ImageLockModeUserInputBuf = 4

PropertyTagFrameDelay = 0x5100


class GdiplusStartupInput(Structure):
    _fields_ = [
        ('GdiplusVersion', c_uint32),
        ('DebugEventCallback', c_void_p),
        ('SuppressBackgroundThread', BOOL),
        ('SuppressExternalCodecs', BOOL)
    ]


class GdiplusStartupOutput(Structure):
    _fields = [
        ('NotificationHookProc', c_void_p),
        ('NotificationUnhookProc', c_void_p)
    ]


class BitmapData(Structure):
    _fields_ = [
        ('Width', c_uint),
        ('Height', c_uint),
        ('Stride', c_int),
        ('PixelFormat', c_int),
        ('Scan0', POINTER(c_byte)),
        ('Reserved', POINTER(c_uint))
    ]


class Rect(Structure):
    _fields_ = [
        ('X', c_int),
        ('Y', c_int),
        ('Width', c_int),
        ('Height', c_int)
    ]


class PropertyItem(Structure):
    _fields_ = [
        ('id', c_uint),
        ('length', c_ulong),
        ('type', c_short),
        ('value', c_void_p)
    ]


INT_PTR = POINTER(INT)
UINT_PTR = POINTER(UINT)


gdiplus.GdipBitmapLockBits.restype = c_int
gdiplus.GdipBitmapLockBits.argtypes = [c_void_p, c_void_p, UINT, c_int, c_void_p]
gdiplus.GdipBitmapUnlockBits.restype = c_int
gdiplus.GdipBitmapUnlockBits.argtypes = [c_void_p, c_void_p]
gdiplus.GdipCloneStringFormat.restype = c_int
gdiplus.GdipCloneStringFormat.argtypes = [c_void_p, c_void_p]
gdiplus.GdipCreateBitmapFromScan0.restype = c_int
gdiplus.GdipCreateBitmapFromScan0.argtypes = [c_int, c_int, c_int, c_int, POINTER(BYTE), c_void_p]
gdiplus.GdipCreateBitmapFromStream.restype = c_int
gdiplus.GdipCreateBitmapFromStream.argtypes = [c_void_p, c_void_p]
gdiplus.GdipCreateFont.restype = c_int
gdiplus.GdipCreateFont.argtypes = [c_void_p, REAL, INT, c_int, c_void_p]
gdiplus.GdipCreateFontFamilyFromName.restype = c_int
gdiplus.GdipCreateFontFamilyFromName.argtypes = [c_wchar_p, c_void_p, c_void_p]
gdiplus.GdipCreateMatrix.restype = None
gdiplus.GdipCreateMatrix.argtypes = [c_void_p]
gdiplus.GdipCreateSolidFill.restype = c_int
gdiplus.GdipCreateSolidFill.argtypes = [c_int, c_void_p]  # ARGB
gdiplus.GdipDisposeImage.restype = c_int
gdiplus.GdipDisposeImage.argtypes = [c_void_p]
gdiplus.GdipDrawString.restype = c_int
gdiplus.GdipDrawString.argtypes = [c_void_p, c_wchar_p, c_int, c_void_p, c_void_p, c_void_p, c_void_p]
gdiplus.GdipGetFamilyName.restype = c_int
gdiplus.GdipGetFamilyName.argtypes = [LONG_PTR, c_wchar_p, c_wchar]
gdiplus.GdipFlush.restype = c_int
gdiplus.GdipFlush.argtypes = [c_void_p, c_int]
gdiplus.GdipGetFontCollectionFamilyCount.restype = c_int
gdiplus.GdipGetFontCollectionFamilyCount.argtypes = [c_void_p, INT_PTR]
gdiplus.GdipGetFontCollectionFamilyList.restype = c_int
gdiplus.GdipGetFontCollectionFamilyList.argtypes = [c_void_p, INT, c_void_p, INT_PTR]
gdiplus.GdipGetImageDimension.restype = c_int
gdiplus.GdipGetImageDimension.argtypes = [c_void_p, POINTER(REAL), POINTER(REAL)]
gdiplus.GdipGetImageGraphicsContext.restype = c_int
gdiplus.GdipGetImageGraphicsContext.argtypes = [c_void_p, c_void_p]
gdiplus.GdipGetImagePixelFormat.restype = c_int
gdiplus.GdipGetImagePixelFormat.argtypes = [c_void_p, c_void_p]
gdiplus.GdipGetPropertyItem.restype = c_int
gdiplus.GdipGetPropertyItem.argtypes = [c_void_p, c_uint, c_uint, c_void_p]
gdiplus.GdipGetPropertyItemSize.restype = c_int
gdiplus.GdipGetPropertyItemSize.argtypes = [c_void_p, c_uint, UINT_PTR]
gdiplus.GdipGraphicsClear.restype = c_int
gdiplus.GdipGraphicsClear.argtypes = [c_void_p, c_int]  # ARGB
gdiplus.GdipImageGetFrameCount.restype = c_int
gdiplus.GdipImageGetFrameCount.argtypes = [c_void_p, c_void_p, UINT_PTR]
gdiplus.GdipImageGetFrameDimensionsCount.restype = c_int
gdiplus.GdipImageGetFrameDimensionsCount.argtypes = [c_void_p, UINT_PTR]
gdiplus.GdipImageGetFrameDimensionsList.restype = c_int
gdiplus.GdipImageGetFrameDimensionsList.argtypes = [c_void_p, c_void_p, UINT]
gdiplus.GdipImageSelectActiveFrame.restype = c_int
gdiplus.GdipImageSelectActiveFrame.argtypes = [c_void_p, c_void_p, UINT]
gdiplus.GdipMeasureString.restype = c_int
gdiplus.GdipMeasureString.argtypes = [c_void_p, c_wchar_p, c_int, c_void_p, c_void_p, c_void_p, c_void_p, INT_PTR, INT_PTR]
gdiplus.GdipNewPrivateFontCollection.restype = c_int
gdiplus.GdipNewPrivateFontCollection.argtypes = [c_void_p]
gdiplus.GdipPrivateAddMemoryFont.restype = c_int
gdiplus.GdipPrivateAddMemoryFont.argtypes = [c_void_p, c_void_p, c_int]
gdiplus.GdipSetPageUnit.restype = c_int
gdiplus.GdipSetPageUnit.argtypes = [c_void_p, c_int]
gdiplus.GdipSetStringFormatFlags.restype = c_int
gdiplus.GdipSetStringFormatFlags.argtypes = [c_void_p, c_int]
gdiplus.GdipSetTextRenderingHint.restype = c_int
gdiplus.GdipSetTextRenderingHint.argtypes = [c_void_p, c_int]
gdiplus.GdipStringFormatGetGenericTypographic.restype = c_int
gdiplus.GdipStringFormatGetGenericTypographic.argtypes = [c_void_p]
gdiplus.GdiplusShutdown.restype = None
gdiplus.GdiplusShutdown.argtypes = [POINTER(ULONG)]
gdiplus.GdiplusStartup.restype = c_int
gdiplus.GdiplusStartup.argtypes = [c_void_p, c_void_p, c_void_p]


class GDIPlusDecoder(ImageDecoder):
    def get_file_extensions(self):
        return ['.bmp', '.gif', '.jpg', '.jpeg', '.exif', '.png', '.tif', '.tiff']

    def get_animation_file_extensions(self):
        # TIFF also supported as a multi-page image; but that's not really an
        # animation, is it?
        return ['.gif']

    def _load_bitmap(self, filename, file):
        data = file.read()

        # Create a HGLOBAL with image data
        hglob = kernel32.GlobalAlloc(GMEM_MOVEABLE, len(data))
        ptr = kernel32.GlobalLock(hglob)
        memmove(ptr, data, len(data))
        kernel32.GlobalUnlock(hglob)

        # Create IStream for the HGLOBAL
        self.stream = pIUnknown()
        ole32.CreateStreamOnHGlobal(hglob, True, byref(self.stream))

        # Load image from stream
        bitmap = c_void_p()
        status = gdiplus.GdipCreateBitmapFromStream(self.stream, byref(bitmap))
        if status != 0:
            self.stream.Release()
            raise ImageDecodeException('GDI+ cannot load %r' % (filename or file))

        return bitmap

    @staticmethod
    def _get_image(bitmap):
        # Get size of image (Bitmap subclasses Image)
        width = REAL()
        height = REAL()
        gdiplus.GdipGetImageDimension(bitmap, byref(width), byref(height))
        width = int(width.value)
        height = int(height.value)

        # Get image pixel format
        pf = c_int()
        gdiplus.GdipGetImagePixelFormat(bitmap, byref(pf))
        pf = pf.value

        # Reverse from what's documented because of Intel little-endianness.
        fmt = 'BGRA'
        if pf == PixelFormat24bppRGB:
            fmt = 'BGR'
        elif pf == PixelFormat32bppRGB:
            pass
        elif pf == PixelFormat32bppARGB:
            pass
        elif pf in (PixelFormat16bppARGB1555, PixelFormat32bppPARGB,
                    PixelFormat64bppARGB, PixelFormat64bppPARGB):
            pf = PixelFormat32bppARGB
        else:
            fmt = 'BGR'
            pf = PixelFormat24bppRGB

        # Lock pixel data in best format
        rect = Rect()
        rect.X = 0
        rect.Y = 0
        rect.Width = width
        rect.Height = height
        bitmap_data = BitmapData()
        gdiplus.GdipBitmapLockBits(bitmap, byref(rect), ImageLockModeRead, pf, byref(bitmap_data))
        
        # Create buffer for RawImage
        buffer = create_string_buffer(bitmap_data.Stride * height)
        memmove(buffer, bitmap_data.Scan0, len(buffer))
        
        # Unlock data
        gdiplus.GdipBitmapUnlockBits(bitmap, byref(bitmap_data))

        return ImageData(width, height, fmt, buffer, -bitmap_data.Stride)

    def _delete_bitmap(self, bitmap):
        # Release image and stream
        gdiplus.GdipDisposeImage(bitmap)
        self.stream.Release()

    def decode(self, filename, file):
        if not file:
            file = open(filename, 'rb')
        bitmap = self._load_bitmap(filename, file)
        image = self._get_image(bitmap)
        self._delete_bitmap(bitmap)
        return image

    def decode_animation(self, filename, file):
        if not file:
            file = open(filename, 'rb')
        bitmap = self._load_bitmap(filename, file)

        dimension_count = c_uint()
        gdiplus.GdipImageGetFrameDimensionsCount(bitmap, byref(dimension_count))
        if dimension_count.value < 1:
            self._delete_bitmap(bitmap)
            raise ImageDecodeException('Image has no frame dimensions')

        # XXX Make sure this dimension is time?
        dimensions = (c_void_p * dimension_count.value)()
        gdiplus.GdipImageGetFrameDimensionsList(bitmap, dimensions, dimension_count.value)

        frame_count = c_uint()
        gdiplus.GdipImageGetFrameCount(bitmap, dimensions, byref(frame_count))

        prop_id = PropertyTagFrameDelay
        prop_size = c_uint()
        gdiplus.GdipGetPropertyItemSize(bitmap, prop_id, byref(prop_size))

        prop_buffer = c_buffer(prop_size.value)
        prop_item = cast(prop_buffer, POINTER(PropertyItem)).contents 
        gdiplus.GdipGetPropertyItem(bitmap, prop_id, prop_size.value, prop_buffer)

        n_delays = prop_item.length // sizeof(c_long)
        delays = cast(prop_item.value, POINTER(c_long * n_delays)).contents

        frames = []

        for i in range(frame_count.value):
            gdiplus.GdipImageSelectActiveFrame(bitmap, dimensions, i)
            image = self._get_image(bitmap)

            delay = delays[i]
            if delay <= 1:
                delay = 10
            frames.append(AnimationFrame(image, delay/100.))

        self._delete_bitmap(bitmap)

        return Animation(frames)


def get_decoders():
    return [GDIPlusDecoder()]


def get_encoders():
    return []


def init():
    token = c_ulong()
    startup_in = GdiplusStartupInput()
    startup_in.GdiplusVersion = 1
    startup_out = GdiplusStartupOutput()
    gdiplus.GdiplusStartup(byref(token), byref(startup_in), byref(startup_out))

    # Shutdown later?
    # gdiplus.GdiplusShutdown(token)


init()



"""
Performance measurement utilities.
"""
import pytest
import time


class PerformanceTimer:
    def __init__(self, max_time):
        self.max_time = max_time
        self.start_time = None

    def __enter__(self):
        self.start_time = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert time.time() - self.start_time < self.max_time


class PerformanceFixture:
    timer = PerformanceTimer


@pytest.fixture
def performance():
    return PerformanceFixture()




"""Windowing and user-interface events.

This module allows applications to create and display windows with an
OpenGL context.  Windows can be created with a variety of border styles
or set fullscreen.

You can register event handlers for keyboard, mouse and window events.
For games and kiosks you can also restrict the input to your windows,
for example disabling users from switching away from the application
with certain key combinations or capturing and hiding the mouse.

Getting started
---------------

Call the Window constructor to create a new window::

    from pyglet.window import Window
    win = Window(width=960, height=540)

Attach your own event handlers::

    @win.event
    def on_key_press(symbol, modifiers):
        # ... handle this event ...

Place drawing code for the window within the `Window.on_draw` event handler::

    @win.event
    def on_draw():
        # ... drawing code ...

Call `pyglet.app.run` to enter the main event loop (by default, this
returns when all open windows are closed)::

    from pyglet import app
    app.run()

Creating a game window
----------------------

Use :py:meth:`~pyglet.window.Window.set_exclusive_mouse` to hide the mouse
cursor and receive relative mouse movement events.  Specify ``fullscreen=True``
as a keyword argument to the :py:class:`~pyglet.window.Window` constructor to
render to the entire screen rather than opening a window::

    win = Window(fullscreen=True)
    win.set_exclusive_mouse()

Working with multiple screens
-----------------------------

By default, fullscreen windows are opened on the primary display (typically
set by the user in their operating system settings).  You can retrieve a list
of attached screens and select one manually if you prefer.  This is useful for
opening a fullscreen window on each screen::

    display = pyglet.display.get_display()
    screens = display.get_screens()
    windows = []
    for screen in screens:
        windows.append(window.Window(fullscreen=True, screen=screen))

Specifying a screen has no effect if the window is not fullscreen.

Specifying the OpenGL context properties
----------------------------------------

Each window has its own context which is created when the window is created.
You can specify the properties of the context before it is created
by creating a "template" configuration::

    from pyglet import gl
    # Create template config
    config = gl.Config()
    config.stencil_size = 8
    config.aux_buffers = 4
    # Create a window using this config
    win = window.Window(config=config)

To determine if a given configuration is supported, query the screen (see
above, "Working with multiple screens")::

    configs = screen.get_matching_configs(config)
    if not configs:
        # ... config is not supported
    else:
        win = window.Window(config=configs[0])

"""
from __future__ import annotations

import sys
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Sequence

import pyglet
import pyglet.window.key
import pyglet.window.mouse
from pyglet import gl
from pyglet.event import EVENT_HANDLE_STATE, EventDispatcher
from pyglet.graphics import shader
from pyglet.math import Mat4
from pyglet.window import event, key

if TYPE_CHECKING:
    from pyglet.display.base import Display, Screen, ScreenMode
    from pyglet.gl import DisplayConfig, Config, Context
    from pyglet.text import Label

_is_pyglet_doc_run = hasattr(sys, 'is_pyglet_doc_run') and sys.is_pyglet_doc_run


class WindowException(Exception):
    """The root exception for all window-related errors."""


class NoSuchDisplayException(WindowException):
    """An exception indicating the requested display is not available."""


class NoSuchConfigException(WindowException):
    """An exception indicating the requested configuration is not available."""


class NoSuchScreenModeException(WindowException):
    """An exception indicating the requested screen resolution could not be met."""


class MouseCursorException(WindowException):
    """The root exception for all mouse cursor-related errors."""


class MouseCursor:
    """An abstract mouse cursor."""

    #: Indicates if the cursor is drawn using OpenGL, or natively.
    gl_drawable: bool = True
    hw_drawable: bool = False

    def draw(self, x: int, y: int) -> None:
        """Abstract render method.

        The cursor should be drawn with the "hot" spot at the given
        coordinates.  The projection is set to the pyglet default (i.e.,
        orthographic in window-space), however no other aspects of the
        state can be assumed.

        Args:
            x:
                X coordinate of the mouse pointer's hot spot.
            y:
                Y coordinate of the mouse pointer's hot spot.

        """


class DefaultMouseCursor(MouseCursor):
    """The default mouse cursor set by the operating system."""
    gl_drawable: bool = False
    hw_drawable: bool = True


class ImageMouseCursor(MouseCursor):
    """A user-defined mouse cursor created from an image.

    Use this class to create your own mouse cursors and assign them
    to windows. Cursors can be drawn by OpenGL, or optionally passed
    to the OS to render natively. There are no restrictions on cursors
    drawn by OpenGL, but natively rendered cursors may have some
    platform limitations (such as color depth, or size). In general,
    reasonably sized cursors will render correctly
    """

    def __init__(self, image: pyglet.image.AbstractImage, hot_x: int = 0, hot_y: int = 0,
                 acceleration: bool = False) -> None:
        """Create a mouse cursor from an image.

        Args:
            image:
                Image to use for the mouse cursor.  It must have a valid ``texture`` attribute.
            hot_x:
                X coordinate of the "hot" spot in the image relative to the image's anchor.
                May be clamped to the maximum image width if acceleration is enabled.
            hot_y:
                Y coordinate of the "hot" spot in the image, relative to the image's anchor.
                May be clamped to the maximum image height if acceleration is enabled.
            acceleration:
                If ``True``, draw the cursor natively instead of using OpenGL.
                The image may be downsampled or color reduced to fit the platform limitations.
        """
        self.texture = image.get_texture()
        self.hot_x = hot_x
        self.hot_y = hot_y
        self.scaling = 1.0

        self.gl_drawable = not acceleration
        self.hw_drawable = acceleration

    def draw(self, x: int, y: int) -> None:
        gl.glEnable(gl.GL_BLEND)
        gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
        self.texture.blit((x - self.hot_x) / self.scaling, (y - self.hot_y) / self.scaling, 0)
        gl.glDisable(gl.GL_BLEND)


def _PlatformEventHandler(data: Any) -> Callable:  # noqa: N802
    """Decorator for platform event handlers.

    Apply giving the platform-specific data needed by the window to associate
    the method with an event.  See platform-specific subclasses of this
    decorator for examples.

    The following attributes are set on the function, which is returned
    otherwise unchanged:

    _platform_event
        True
    _platform_event_data
        List of data applied to the function (permitting multiple decorators
        on the same method).
    """

    def _event_wrapper(f: Callable) -> Callable:
        f._platform_event = True  # noqa: SLF001
        if not hasattr(f, '_platform_event_data'):
            f._platform_event_data = []  # noqa: SLF001
        f._platform_event_data.append(data)  # noqa: SLF001
        return f

    return _event_wrapper


def _ViewEventHandler(f: Callable) -> Callable:  # noqa: N802
    f._view = True  # noqa: SLF001
    return f


class _WindowMetaclass(type):
    """Sets the _platform_event_names class variable on the window subclass."""

    def __init__(cls: type[_WindowMetaclass], name: str, bases: tuple, dct: dict) -> None:
        cls._platform_event_names = set()
        for base in bases:
            if hasattr(base, '_platform_event_names'):
                cls._platform_event_names.update(base._platform_event_names)
        for name, func in dct.items():
            if hasattr(func, '_platform_event'):
                cls._platform_event_names.add(name)
        super().__init__(name, bases, dct)


class BaseWindow(EventDispatcher, metaclass=_WindowMetaclass):
    """Platform-independent application window.

    A window is a "heavyweight" object occupying operating system resources.
    The "client" or "content" area of a window is filled entirely with
    an OpenGL viewport.  Applications have no access to operating system
    widgets or controls; all rendering must be done via OpenGL.

    Windows may appear as floating regions or can be set to fill an entire
    screen (fullscreen).  When floating, windows may appear borderless or
    decorated with a platform-specific frame (including, for example, the
    title bar, minimize and close buttons, resize handles, and so on).

    While it is possible to set the location of a window, it is recommended
    that applications allow the platform to place it according to local
    conventions.  This will ensure it is not obscured by other windows,
    and appears on an appropriate screen for the user.

    To render into a window, you must first call its :py:meth:`.switch_to`
    method to make it the active OpenGL context. If you use only one
    window in your application, you can skip this step as it will always
    be the active context.
    """

    # Filled in by metaclass with the names of all methods on this (sub)class
    # that are platform event handlers.
    _platform_event_names: set[_PlatformEventHandler] = set()  # noqa: RUF012

    #: The default window style.
    WINDOW_STYLE_DEFAULT: None = None
    #: The window style for pop-up dialogs.
    WINDOW_STYLE_DIALOG: str = 'dialog'
    #: The window style for tool windows.
    WINDOW_STYLE_TOOL: str = 'tool'
    #: A window style without any decoration.
    WINDOW_STYLE_BORDERLESS: str = 'borderless'
    #: A window style for transparent, interactable windows
    WINDOW_STYLE_TRANSPARENT: str = 'transparent'
    #: A window style for transparent, topmost, click-through-able overlays
    WINDOW_STYLE_OVERLAY: str = 'overlay'

    #: The default mouse cursor.
    CURSOR_DEFAULT = None
    #: A crosshair mouse cursor.
    CURSOR_CROSSHAIR: str = 'crosshair'
    #: A pointing hand mouse cursor.
    CURSOR_HAND: str = 'hand'
    #: A "help" mouse cursor; typically a question mark and an arrow.
    CURSOR_HELP: str = 'help'
    #: A mouse cursor indicating that the selected operation is not permitted.
    CURSOR_NO: str = 'no'
    #: A mouse cursor indicating the element can be resized.
    CURSOR_SIZE: str = 'size'
    #: A mouse cursor indicating the element can be resized from the top
    #: border.
    CURSOR_SIZE_UP: str = 'size_up'
    #: A mouse cursor indicating the element can be resized from the
    #: upper-right corner.
    CURSOR_SIZE_UP_RIGHT: str = 'size_up_right'
    #: A mouse cursor indicating the element can be resized from the right
    #: border.
    CURSOR_SIZE_RIGHT: str = 'size_right'
    #: A mouse cursor indicating the element can be resized from the lower-right
    #: corner.
    CURSOR_SIZE_DOWN_RIGHT: str = 'size_down_right'
    #: A mouse cursor indicating the element can be resized from the bottom
    #: border.
    CURSOR_SIZE_DOWN: str = 'size_down'
    #: A mouse cursor indicating the element can be resized from the lower-left
    #: corner.
    CURSOR_SIZE_DOWN_LEFT: str = 'size_down_left'
    #: A mouse cursor indicating the element can be resized from the left
    #: border.
    CURSOR_SIZE_LEFT: str = 'size_left'
    #: A mouse cursor indicating the element can be resized from the upper-left
    #: corner.
    CURSOR_SIZE_UP_LEFT: str = 'size_up_left'
    #: A mouse cursor indicating the element can be resized vertically.
    CURSOR_SIZE_UP_DOWN: str = 'size_up_down'
    #: A mouse cursor indicating the element can be resized horizontally.
    CURSOR_SIZE_LEFT_RIGHT: str = 'size_left_right'
    #: A text input mouse cursor (I-beam).
    CURSOR_TEXT: str = 'text'
    #: A "wait" mouse cursor; typically an hourglass or watch.
    CURSOR_WAIT: str = 'wait'
    #: The "wait" mouse cursor combined with an arrow.
    CURSOR_WAIT_ARROW: str = 'wait_arrow'

    #: True if the user has attempted to close the window.
    #:
    #: :deprecated: Windows are closed immediately by the default
    #:      :py:meth:`~pyglet.window.Window.on_close` handler when `pyglet.app.event_loop` is being
    #:      used.
    has_exit: bool = False

    #: Window display contents validity.  The :py:mod:`pyglet.app` event loop
    #: examines every window each iteration and only dispatches the :py:meth:`~pyglet.window.Window.on_draw`
    #: event to windows that have `invalid` set.  By default, windows always
    #: have `invalid` set to ``True``.
    #:
    #: You can prevent redundant redraws by setting this variable to ``False``
    #: in the window's :py:meth:`~pyglet.window.Window.on_draw` handler, and setting it to True again in
    #: response to any events that actually do require a window contents
    #: update.
    #:
    #: :type: bool
    #:
    #: .. versionadded:: 1.1
    invalid: bool = True

    # Instance variables accessible only via properties
    _dpi: int = 96
    _width: int | None = None
    _height: int | None = None
    _caption: str | None = None
    _resizable: bool = False
    _style: str | None = WINDOW_STYLE_DEFAULT
    _fullscreen: bool = False
    _visible: bool = False
    _vsync: bool = False
    _file_drops: bool = False
    _screen: Screen | None = None
    _config: DisplayConfig | None = None
    _context: Context | None = None
    _projection_matrix: Mat4 = pyglet.math.Mat4()
    _view_matrix: Mat4 = pyglet.math.Mat4()
    _viewport: tuple[int, int, int, int] = 0, 0, 0, 0

    # Used to restore window size and position after fullscreen
    _windowed_size: tuple[int, int] | None = None
    _windowed_location: tuple[int, int] | None = None

    _minimum_size: tuple[int, int] | None = None
    _maximum_size: tuple[int, int] | None = None

    _keyboard_exclusive: bool = False

    _shadow: bool = False

    # Subclasses should update these after relevant events
    _mouse_cursor: MouseCursor | ImageMouseCursor = DefaultMouseCursor()
    _mouse_x: int = 0
    _mouse_y: int = 0
    _mouse_visible: bool = True
    _mouse_exclusive: bool = False
    _mouse_in_window: bool = False

    _event_queue = None
    _enable_event_queue: bool = True  # overridden by EventLoop.
    _allow_dispatch_event: bool = False  # controlled by dispatch_events stack frame

    # Class attributes
    _default_width: int = 1280
    _default_height: int = 720

    _requested_width: int
    _requested_height: int

    # Create a default ShaderProgram, so the Window instance can
    # update the `WindowBlock` UBO shared by all default shaders.
    _default_vertex_source = """#version 150 core
        in vec4 position;

        uniform WindowBlock
        {
            mat4 projection;
            mat4 view;
        } window;

        void main()
        {
            gl_Position = window.projection * window.view * position;
        }
    """
    _default_fragment_source = """#version 150 core
        out vec4 color;

        void main()
        {
            color = vec4(1.0, 0.0, 0.0, 1.0);
        }
    """

    def __init__(self,
                 width: int | None = None,
                 height: int | None = None,
                 caption: str | None = None,
                 resizable: bool = False,
                 style: str | None = WINDOW_STYLE_DEFAULT,
                 fullscreen: bool = False,
                 visible: bool = True,
                 vsync: bool = True,
                 file_drops: bool = False,
                 display: Display | None = None,
                 screen: Screen | None = None,
                 config: Config | None = None,
                 context: Context | None = None,
                 mode: ScreenMode | None = None) -> None:
        """Create a window.

        All parameters are optional, and reasonable defaults are assumed
        where they are not specified.

        The ``display``, ``screen``, ``config`` and ``context`` parameters form
        a hierarchy of control: there is no need to specify more than
        one of these.  For example, if you specify ``screen`` the ``display``
        will be inferred, and a default ``config`` and ``context`` will be
        created.

        ``config`` is a special case; it can be a template created by the
        user specifying the attributes desired, or it can be a complete
        ``config`` as returned from :py:meth:`~pyglet.display.Screen.get_matching_configs` or similar.

        The context will be active as soon as the window is created, as if
        :py:meth:`~pyglet.window.Window.switch_to` was just called.

        Args:
            width:
                Width of the window, in pixels.  Defaults to 960, or the screen width if ``fullscreen`` is True.
            height:
                Height of the window, in pixels.  Defaults to 540, or the screen height if ``fullscreen`` is True.
            caption:
                Initial caption (title) of the window.  Defaults to ``sys.argv[0]``.
            resizable:
                If True, the window will be resizable.  Defaults to False.
            style:
                One of the ``WINDOW_STYLE_*`` constants specifying the border style of the window.
            fullscreen:
                If True, the window will cover the entire screen rather than floating.  Defaults to False.
            visible:
                Determines if the window is visible immediately after
                creation.  Defaults to True.  Set this to False if you
                would like to change attributes of the window before
                having it appear to the user.
            vsync:
                If True, buffer flips are synchronised to the primary screen's
                vertical retrace, eliminating flicker.
            file_drops:
                If True, the Window will accept files being dropped into it and call the ``on_file_drop`` event.
            display:
                The display device to use.  Useful only under X11.
            screen:
                The screen to use, if in fullscreen.
            config:
                Either a template from which to create a complete config, or a complete config.
            context:
                The context to attach to this window.  The context must not already be attached to another window.
            mode:
                The screen will be switched to this mode if `fullscreen` is
                True.  If None, an appropriate mode is selected to accommodate ``width`` and ``height``.

        """
        EventDispatcher.__init__(self)
        self._event_queue = []

        if not display:
            display = pyglet.display.get_display()

        if not screen:
            screen = display.get_default_screen()

        if not config:
            alpha_size = None
            transparent_fb = False
            # Override config settings if intention is transparency.
            if style in ('transparent', 'overlay'):
                # Ensure the framebuffer is large enough to support transparency.
                alpha_size = 8
                transparent_fb = True

            for template_config in [
                gl.Config(double_buffer=True, depth_size=24, major_version=3, minor_version=3,
                          alpha_size=alpha_size, transparent_framebuffer=transparent_fb),
                gl.Config(double_buffer=True, depth_size=16, major_version=3, minor_version=3,
                          alpha_size=alpha_size, transparent_framebuffer=transparent_fb),
                None,
            ]:
                try:
                    config = screen.get_best_config(template_config)
                    break
                except NoSuchConfigException:
                    pass
            if not config:
                msg = 'No standard config is available.'
                raise NoSuchConfigException(msg)
        else:
            if style in ('transparent', 'overlay'):
                config.alpha_size = 8
                config.transparent_framebuffer = True

        if not config.is_complete():
            config = screen.get_best_config(config)

        if not context:
            context = config.create_context(gl.current_context)

        # Set these in reverse order as above, to ensure we get user preference
        self._context = context
        self._config = self._context.config

        # XXX deprecate config's being screen-specific
        if hasattr(self._config, 'screen'):
            self._screen = self._config.screen
        else:
            self._screen = screen
        self._display = self._screen.display

        if fullscreen:
            if width is None and height is None:
                self._windowed_size = self._default_width, self._default_height
            width, height = self._set_fullscreen_mode(mode, width, height)
            if not self._windowed_size:
                self._windowed_size = width, height
        else:
            if width is None:
                width = self._default_width
            if height is None:
                height = self._default_height

        self._width = width
        self._height = height
        self._requested_width = width
        self._requested_height = height

        self._resizable = resizable
        self._fullscreen = fullscreen
        self._style = style
        if pyglet.options['vsync'] is not None:
            self._vsync = pyglet.options['vsync']
        else:
            self._vsync = vsync

        self._file_drops = file_drops
        self._caption = caption or sys.argv[0]

        from pyglet import app
        app.windows.add(self)
        self._create()

        self.switch_to()

        self._create_projection()

        if visible:
            self.set_visible(True)
            self.activate()

    def _create_projection(self) -> None:
        self._default_program = shader.ShaderProgram(
            shader.Shader(self._default_vertex_source, 'vertex'),
            shader.Shader(self._default_fragment_source, 'fragment'))

        self.ubo = self._default_program.uniform_blocks['WindowBlock'].create_ubo()

        self._viewport = 0, 0, *self.get_framebuffer_size()

        width, height = self.get_size()
        self.view = Mat4()
        self.projection = Mat4.orthogonal_projection(0, width, 0, height, -8192, 8192)

    def __del__(self) -> None:
        # Always try to clean up the window when it is dereferenced.
        # Makes sure there are no dangling pointers or memory leaks.
        # If the window is already closed, pass silently.
        try:  # noqa: SIM105
            self.close()
        except:  # XXX  Avoid a NoneType error if already closed.  # noqa: E722, S110
            pass

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}=(width={self.width}, height={self.height})'

    @abstractmethod
    def _create(self) -> None:
        ...

    @abstractmethod
    def _recreate(self, changes: Sequence[str]) -> None:
        """Recreate the window with current attributes.

        Args:
            changes:
                Sequence of attribute names that were changed since the last
                ``_create`` or ``_recreate``.  For example, ``['fullscreen']``
                is given if the window is to be toggled to or from fullscreen.
        """

    # Public methods (sort alphabetically):
    @abstractmethod
    def activate(self) -> None:
        """Attempt to restore keyboard focus to the window.

        Depending on the window manager or operating system, this may not
        be successful.  For example, on Windows XP an application is not
        allowed to "steal" focus from another application.  Instead, the
        window's taskbar icon will flash, indicating it requires attention.
        """

    @staticmethod
    def clear() -> None:
        """Clear the window.

        This is a convenience method for clearing the color and depth
        buffer.  The window must be the active context (see
        :py:meth:`.switch_to`).
        """
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)

    def close(self) -> None:
        """Close the window.

        After closing the window, the GL context will be invalid.  The
        window instance cannot be reused once closed. To reuse windows,
        see :py:meth:`.set_visible` instead.

        The :py:meth:`pyglet.app.EventLoop.on_window_close` event is
        dispatched by the :py:attr:`pyglet.app.event_loop` when this method
        is called.
        """
        from pyglet import app
        if not self._context:
            return
        app.windows.remove(self)
        self._context.destroy()
        self._config = None
        self._context = None
        if app.event_loop:
            app.event_loop.dispatch_event('on_window_close', self)
        self._event_queue = []

    def dispatch_event(self, *args: Any) -> None:
        if not self._enable_event_queue or self._allow_dispatch_event:
            super().dispatch_event(*args)
        else:
            self._event_queue.append(args)

    @abstractmethod
    def dispatch_events(self) -> None:
        """Poll the operating system event queue for new events and call attached event handlers.

        This method is provided for legacy applications targeting pyglet 1.0,
        and advanced applications that must integrate their event loop
        into another framework.

        Typical applications should use :py:func:`pyglet.app.run`.
        """

    def draw(self, dt: float) -> None:
        """Redraw the Window contents.

        This method will first call the :py:meth:`~pyglet.window.Window.`switch_to`
        method to make the GL context current. It then dispatches the
        :py:meth:`~pyglet.window.Window.on_draw` and
        :py:meth:`~pyglet.window.Window.on_refresh`
        events. Finally, it calls the :py:meth:`~pyglet.window.Window.flip`
        method to swap the front and back OpenGL buffers.
        """
        self.switch_to()
        self.dispatch_event('on_draw')
        self.dispatch_event('on_refresh', dt)
        self.flip()

    def draw_mouse_cursor(self) -> None:
        """Draw the custom mouse cursor.

        If the current mouse cursor has ``drawable`` set, this method
        is called before the buffers are flipped to render it.

        There is little need to override this method; instead, subclass
        :py:class:`MouseCursor` and provide your own
        :py:meth:`~MouseCursor.draw` method.
        """
        # Draw mouse cursor if set and visible.

        if self._mouse_cursor.gl_drawable and self._mouse_visible and self._mouse_in_window:
            # TODO: consider projection differences
            self._mouse_cursor.draw(self._mouse_x, self._mouse_y)

    @abstractmethod
    def flip(self) -> None:
        """Swap the OpenGL front and back buffers.

        Call this method on a double-buffered window to update the
        visible display with the back buffer. Windows are
        double-buffered by default unless you turn this feature off.

        The contents of the back buffer are undefined after this operation.

        The default :py:attr:`~pyglet.app.event_loop` automatically
        calls this method after the window's
        :py:meth:`~pyglet.window.Window.on_draw` event.
        """

    def get_framebuffer_size(self) -> tuple[int, int]:
        """Return the size in actual pixels of the Window framebuffer.

        When using HiDPI screens, the size of the Window's framebuffer
        can be higher than that of the Window size requested. If you
        are performing operations that require knowing the actual number
        of pixels in the window, this method should be used instead of
        :py:func:`Window.get_size()`. For example, setting the Window
        projection or setting the glViewport size.

        Returns:
            The width and height of the Window's framebuffer, in pixels.
        """
        return self.get_size()

    @abstractmethod
    def get_location(self) -> tuple[int, int]:
        """Return the current position of the window.

        Returns:
             The distances of the left and top edges from their respective edges on the virtual desktop, in pixels.
        """

    def get_pixel_ratio(self) -> float:
        """Return the framebuffer/window size ratio.

        Some platforms and/or window systems support subpixel scaling,
        making the framebuffer size larger than the window size.
        Retina screens on OS X and Gnome on Linux are some examples.

        On a Retina systems the returned ratio would usually be 2.0 as a
        window of size 500 x 500 would have a framebuffer of 1000 x 1000.
        Fractional values between 1.0 and 2.0, as well as values above
        2.0 may also be encountered.

        :deprecated: Use `Window.scale`.
        """
        return self.scale

    def get_size(self) -> tuple[int, int]:
        """Return the current size of the window.

        This does not include the windows' border or title bar.

        Returns:
            The width and height of the window, in pixels.
        """
        return self._width, self._height

    def get_requested_size(self) -> tuple[int, int]:
        """Return the size of the window without any scaling taken into effect.

        This does not include the windows' border or title bar.

        Returns:
            The width and height of the window, in pixels.
        """
        return self._requested_width, self._requested_height

    def get_system_mouse_cursor(self, name: str) -> MouseCursor:
        """Obtain a system mouse cursor.

        Use :py:meth:`~pyglet.window.Window.set_mouse_cursor` to make the cursor returned by this method
        active.  The names accepted by this method are the ``CURSOR_*``
        constants defined on this class.
        """
        raise NotImplementedError

    def get_clipboard_text(self) -> str:
        """Access the system clipboard and attempt to retrieve text.

        Returns:
             A string from the clipboard. String will be empty if no text found.
        """
        raise NotImplementedError

    def set_clipboard_text(self, text: str) -> None:
        """Access the system clipboard and set a text string as the clipboard data.

        This will clear the existing clipboard.
        """
        raise NotImplementedError

    def set_mouse_passthrough(self, state: bool) -> None:
        """Set whether the operating system will ignore mouse input from this window.

        Behavior may differ across operating systems. This is typically used in window overlays with
        transparent frame buffers.

        Args:
            state:
                ``True`` will allow mouse input to pass through the window to anything behind it. Otherwise, ``False``
                allows the window to accept focus again.

        .. versionadded:: 2.1.8
        """
        raise NotImplementedError

    @abstractmethod
    def minimize(self) -> None:
        """Minimize the window."""

    @abstractmethod
    def maximize(self) -> None:
        """Maximize the window.

        The behaviour of this method is somewhat dependent on the user's
        display setup.  On a multi-monitor system, the window may maximize
        to either a single screen or the entire virtual desktop.
        """

    def on_close(self) -> None:
        """Default on_close handler."""
        self.has_exit = True
        from pyglet import app
        if app.event_loop.is_running:
            self.close()

    def on_key_press(self, symbol: int, modifiers: int) -> EVENT_HANDLE_STATE:
        """Default on_key_press handler."""
        if symbol == key.ESCAPE and not (modifiers & ~(key.MOD_NUMLOCK |
                                                       key.MOD_CAPSLOCK |
                                                       key.MOD_SCROLLLOCK)):
            self.dispatch_event('on_close')

    def _on_internal_resize(self, width: int, height: int) -> None:
        framebuffer_size = self.get_framebuffer_size()
        gl.glViewport(0, 0, max(framebuffer_size[0], 1), max(framebuffer_size[1], 1))
        w, h = self.get_size()
        self.projection = Mat4.orthogonal_projection(0, max(w, 1), 0, max(h, 1), -8192, 8192)
        self.dispatch_event('on_resize', w, h)

    def _on_internal_scale(self, scale: float, dpi: int) -> None:
        framebuffer_size = self.get_framebuffer_size()
        gl.glViewport(0, 0, max(framebuffer_size[0], 1), max(framebuffer_size[1], 1))
        w, h = self.get_size()
        self.projection = Mat4.orthogonal_projection(0, max(w, 1), 0, max(h, 1), -8192, 8192)
        self._mouse_cursor.scaling = self._get_mouse_scale()
        self.dispatch_event('on_scale', scale, dpi)

    def on_resize(self, width: int, height: int) -> EVENT_HANDLE_STATE:
        """A default resize event handler.

        This default handler updates the GL viewport to cover the entire
        window. The bottom-left corner is (0, 0) and the top-right
        corner is the width and height of the window's framebuffer.
        In addition, the projection matrix is set to an orthogonal
        projection based on the same dimensions.
        """

    def on_scale(self, scale: float, dpi: int) -> EVENT_HANDLE_STATE:
        """A default scale event handler.

        This default handler is called if the screen or system's DPI changes
        during runtime.
        """

    @abstractmethod
    def set_caption(self, caption: str) -> None:
        """Set the window's caption.

        The caption appears in the titlebar of the window, if it has one,
        and in the taskbar on Windows and many X11 window managers.
        """

    def set_fullscreen(self, fullscreen: bool = True, screen: Screen | None = None, mode: ScreenMode | None = None,
                       width: int | None = None, height: int | None = None) -> None:
        """Toggle to or from fullscreen.

        After toggling fullscreen, the GL context should have retained its
        state and objects, however the buffers will need to be cleared and
        redrawn.

        If ``width`` and ``height`` are specified and ``fullscreen`` is True, the
        screen may be switched to a different resolution that most closely
        matches the given size.  If the resolution doesn't match exactly,
        a higher resolution is selected and the window will be centered
        within a black border covering the rest of the screen.

        Args:
            fullscreen:
                True if the window should be made fullscreen, False if it
                should be windowed.
            screen:
                If not None and fullscreen is True, the window is moved to the
                given screen.  The screen must belong to the same display as
                the window.
            mode:
                The screen will be switched to the given mode.  The mode must
                have been obtained by enumerating :py:meth:`~pyglet.display.Screen.get_modes`.  If
                None, an appropriate mode will be selected from the given
                ``width`` and ``height``.
            width: int
                Optional width of the window.  If unspecified, defaults to the
                previous window size when windowed, or the screen size if
                fullscreen.

                .. versionadded:: 1.2
            height:
                Optional height of the window.  If unspecified, defaults to
                the previous window size when windowed, or the screen size if
                fullscreen.

                .. versionadded:: 1.2
        """
        if (fullscreen == self._fullscreen and
                (screen is None or screen is self._screen) and
                (width is None or width == self._width) and
                (height is None or height == self._height)):
            return

        if not self._fullscreen:
            # Save windowed size
            self._windowed_size = self.get_size()
            self._windowed_location = self.get_location()

        if fullscreen and screen is not None:
            assert screen.display is self.display
            self._screen = screen

        self._fullscreen = fullscreen
        if self._fullscreen:
            self._width, self._height = self._set_fullscreen_mode(mode, width, height)
        else:
            self.screen.restore_mode()

            self._width, self._height = self._windowed_size
            if width is not None:
                self._width = width
            if height is not None:
                self._height = height

        self._recreate(['fullscreen'])

        if not self._fullscreen and self._windowed_location:
            # Restore windowed location.
            self.set_location(*self._windowed_location)

    def _set_fullscreen_mode(self, mode: ScreenMode, width: int, height: int) -> tuple[int, int]:
        if mode is not None:
            self.screen.set_mode(mode)
            if width is None:
                width = self.screen.width
            if height is None:
                height = self.screen.height
        elif width is not None or height is not None:
            if width is None:
                width = 0
            if height is None:
                height = 0
            mode = self.screen.get_closest_mode(width, height)
            if mode is not None:
                self.screen.set_mode(mode)
            elif self.screen.get_modes():
                # Only raise exception if mode switching is at all possible.
                msg = f'No mode matching {width}x{height}'
                raise NoSuchScreenModeException(msg)
        else:
            width = self.screen.width
            height = self.screen.height
        return width, height

    def set_minimum_size(self, width: int, height: int) -> None:
        """Set the minimum size of the window.

        Once set, the user will not be able to resize the window smaller
        than the given dimensions.  There is no way to remove the
        minimum size constraint on a window (but you could set it to 0,0).

        The behaviour is undefined if the minimum size is set larger than
        the current size of the window.

        The window size does not include the border or title bar.

        Args:
            width:
                Minimum width of the window, in pixels.
            height:
                Minimum height of the window, in pixels.

        """
        if width < 1 or height < 1:
            msg = 'Width and height must be positive integers.'
            raise ValueError(msg)

        self._minimum_size = width, height

    def set_maximum_size(self, width: int, height: int) -> None:
        """Set the maximum size of the window.

        Once set, the user will not be able to resize the window larger
        than the given dimensions.  There is no way to remove the
        maximum size constraint on a window (but you could set it to a large
        value).

        The behaviour is undefined if the maximum size is set smaller than
        the current size of the window.

        The window size does not include the border or title bar.

        Args:
            width:
                Maximum width of the window, in pixels.
            height:
                Maximum height of the window, in pixels.

        """
        if width < 1 or height < 1:
            msg = 'Width and height must be positive integers'
            raise ValueError(msg)

        self._maximum_size = width, height

    def set_size(self, width: int, height: int) -> None:
        """Resize the window.

        The behaviour is undefined if the window is not resizable, or if
        it is currently fullscreen.

        The window size does not include the border or title bar.
        """
        if self._fullscreen:
            msg = 'Cannot set size of fullscreen window.'
            raise WindowException(msg)
        if width < 1 or height < 1:
            msg = 'width and height must be positive integers'
            raise ValueError(msg)

        self._width, self._height = width, height
        self._requested_width, self._requested_height = width, height

    @abstractmethod
    def set_location(self, x: int, y: int) -> None:
        """Set the position of the window.

        Args:
            x:
                Distance of the left edge of the window from the left edge of the virtual desktop, in pixels.
            y:
                Distance of the top edge of the window from the top edge of the virtual desktop, in pixels.
        """

    def set_visible(self, visible: bool = True) -> None:
        """Show or hide the window."""
        self._visible = visible

    def set_vsync(self, vsync: bool) -> None:
        """Enable or disable vertical sync control.

        When enabled, this option ensures flips from the back to the front
        buffer are performed only during the vertical retrace period of the
        primary display.  This can prevent "tearing" or flickering when
        the buffer is updated in the middle of a video scan.

        Note that LCD monitors have an analogous time in which they are not
        reading from the video buffer; while it does not correspond to
        a vertical retrace it has the same effect.

        Also note that with multi-monitor systems the secondary monitor
        cannot be synchronised to, so tearing and flicker cannot be avoided
        when the window is positioned outside of the primary display.
        """
        self._vsync = vsync

    def set_mouse_visible(self, visible: bool = True) -> None:
        """Show or hide the mouse cursor.

        The mouse cursor will only be hidden while it is positioned within
        this window.  Mouse events will still be processed as usual.
        """
        self._mouse_visible = visible
        self.set_mouse_platform_visible()

    def set_mouse_platform_visible(self, platform_visible: bool | None = None) -> None:
        """Set the platform-drawn mouse cursor visibility.

        This is called automatically after changing the mouse cursor or exclusive mode.

        Applications should not normally need to call this method.

        :see: :py:meth:`~pyglet.window.Window.set_mouse_visible` instead.

        Args:
            platform_visible:
                If ``None``, sets platform visibility to the required visibility
                for the current exclusive mode and cursor type.  Otherwise,
                a bool value will override and force a visibility.

        """
        raise NotImplementedError

    def set_mouse_cursor(self, cursor: MouseCursor | None = None) -> None:
        """Change the appearance of the mouse cursor.

        The appearance of the mouse cursor is only changed while it is
        within this window.

        Args:
            cursor:
                The cursor to set, or ``None`` to restore the default cursor.

        """
        if cursor is None:
            cursor = DefaultMouseCursor()
        self._mouse_cursor = cursor
        self._mouse_cursor.scaling = self._get_mouse_scale()
        self.set_mouse_platform_visible()

    def _get_mouse_scale(self):
        """The mouse scale factoring in the DPI.

        On Mac, this is always 1.0.
        """
        return self.scale

    def set_exclusive_mouse(self, exclusive: bool = True) -> None:
        """Hide the mouse cursor and direct all mouse events to this window.

        When enabled, this feature prevents the mouse leaving the window.  It
        is useful for certain styles of games that require complete control of
        the mouse.  The position of the mouse as reported in subsequent events
        is meaningless when exclusive mouse is enabled; you should only use
        the relative motion parameters ``dx`` and ``dy``.
        """
        self._mouse_exclusive = exclusive

    def set_exclusive_keyboard(self, exclusive: bool = True) -> None:
        """Prevent the user from switching away from this window using keyboard accelerators.

        When enabled, this feature disables certain operating-system specific
        key combinations such as Alt+Tab (Command+Tab on OS X).  This can be
        useful in certain kiosk applications, it should be avoided in general
        applications or games.
        """
        self._keyboard_exclusive = exclusive

    def set_icon(self, *images: pyglet.image.AbstractImage) -> None:
        """Set the window icon.

        If multiple images are provided, one with an appropriate size
        will be selected (if the correct size is not provided, the image
        will be scaled).

        Useful sizes to provide are 16x16, 32x32, 64x64 (Mac only) and
        128x128 (Mac only).
        """

    @abstractmethod
    def switch_to(self) -> None:
        """Make this window the current OpenGL rendering context.

        Only one OpenGL context can be active at a time. This method
        sets the current window context as the active one.

        In most cases, you should use this method instead of directly
        calling :py:meth:`~pyglet.gl.Context.set_current`. The latter
        will not perform platform-specific state management tasks for
        you.
        """

    # Attributes (sort alphabetically):
    @property
    def caption(self) -> str:
        """The window caption (title). Read-only."""
        return self._caption

    @property
    def resizeable(self) -> bool:
        """True if the window is resizable. Read-only."""
        return self._resizable

    @property
    def style(self) -> str:
        """The window style; one of the ``WINDOW_STYLE_*`` constants. Read-only."""
        return self._style

    @property
    def fullscreen(self) -> bool:
        """True if the window is currently fullscreen. Read-only."""
        return self._fullscreen

    @property
    def visible(self) -> bool:
        """True if the window is currently visible. Read-only."""
        return self._visible

    @property
    def vsync(self) -> bool:
        """True if buffer flips are synchronised to the screen's vertical retrace. Read-only."""
        return self._vsync

    @property
    def display(self) -> Display:
        """The display this window belongs to.  Read-only."""
        return self._display

    @property
    def screen(self) -> Screen:
        """The screen this window is fullscreen in.  Read-only."""
        return self._screen

    @property
    def config(self) -> DisplayConfig:
        """A GL config describing the context of this window.  Read-only."""
        return self._config

    @property
    def context(self) -> Context:
        """The OpenGL context attached to this window.  Read-only."""
        return self._context

    # These are the only properties that can be set
    @property
    def width(self) -> int:
        """The width of the window, in pixels.  Read-write."""
        return self.get_size()[0]

    @width.setter
    def width(self, new_width: int) -> None:
        self.set_size(new_width, self.height)

    @property
    def height(self) -> int:
        """The height of the window, in pixels.  Read-write."""
        return self.get_size()[1]

    @height.setter
    def height(self, new_height: int) -> None:
        self.set_size(self.width, new_height)

    @property
    def scale(self) -> float:
        """The scale of the window factoring in DPI.

        Read only.
        """
        if pyglet.options.dpi_scaling != "real":
            return self._dpi / 96

        return 1.0

    @property
    def dpi(self) -> int:
        """DPI values of the Window.

        Read only.
        """
        return self._dpi

    @property
    def size(self) -> tuple[int, int]:
        """The size of the window. Read-Write."""
        return self.get_size()

    @size.setter
    def size(self, new_size: Sequence[int, int]) -> None:
        self.set_size(*new_size)

    @property
    def aspect_ratio(self) -> float:
        """The aspect ratio of the window. Read-Only."""
        w, h = self.get_size()
        return w / h

    @property
    def projection(self) -> Mat4:
        """The OpenGL window projection matrix. Read-write.

        This matrix is used to transform vertices when using any of the built-in
        drawable classes. `view` is done first, then `projection`.

        The default projection matrix is orthographic (2D),
        but a custom :py:class:`~pyglet.math.Mat4` instance
        can be set. Alternatively, you can supply a flat
        tuple of 16 values.

        (2D), but can be changed to any 4x4 matrix desired.
        :see: :py:class:`~pyglet.math.Mat4`.
        """
        return self._projection_matrix

    @projection.setter
    def projection(self, matrix: Mat4) -> None:

        with self.ubo as window_block:
            window_block.projection[:] = matrix

        self._projection_matrix = matrix

    @property
    def view(self) -> Mat4:
        """The OpenGL window view matrix. Read-write.

        This matrix is used to transform vertices when using any of the built-in
        drawable classes. `view` is done first, then `projection`.

        The default view is an identity matrix, but a custom
        :py:class:`~pyglet.math.Mat4` instance can be set.
        Alternatively, you can supply a flat tuple of 16 values.
        """
        return self._view_matrix

    @view.setter
    def view(self, matrix: Mat4) -> None:

        with self.ubo as window_block:
            window_block.view[:] = matrix

        self._view_matrix = matrix

    @property
    def viewport(self) -> tuple[int, int, int, int]:
        """The Window viewport.

        The Window viewport, expressed as (x, y, width, height).
        """
        return self._viewport

    @viewport.setter
    def viewport(self, values: tuple[int, int, int, int]) -> None:
        self._viewport = values
        pr = self.scale
        x, y, w, h = values
        pyglet.gl.glViewport(int(x * pr), int(y * pr), int(w * pr), int(h * pr))

    # If documenting, show the event methods.  Otherwise, leave them out
    # as they are not really methods.
    if _is_pyglet_doc_run:
        def on_activate(self) -> EVENT_HANDLE_STATE:
            """The window was activated.

            This event can be triggered by clicking on the title bar, bringing
            it to the foreground; or by some platform-specific method.

            When a window is "active" it has the keyboard focus.

            :event:
            """

        def on_close(self) -> EVENT_HANDLE_STATE:
            """The user attempted to close the window.

            This event can be triggered by clicking on the "X" control box in
            the window title bar, or by some other platform-dependent manner.

            The default handler sets `has_exit` to ``True``.  In pyglet 1.1, if
            `pyglet.app.event_loop` is being used, `close` is also called,
            closing the window immediately.

            :event:
            """

        def on_context_lost(self) -> EVENT_HANDLE_STATE:
            """The window's GL context was lost.

            When the context is lost no more GL methods can be called until it
            is recreated.  This is a rare event, triggered perhaps by the user
            switching to an incompatible video mode.  When it occurs, an
            application will need to reload all objects (display lists, texture
            objects, shaders) as well as restore the GL state.

            :event:
            """

        def on_context_state_lost(self) -> EVENT_HANDLE_STATE:
            """The state of the window's GL context was lost.

            pyglet may sometimes need to recreate the window's GL context if
            the window is moved to another video device, or between fullscreen
            or windowed mode.  In this case it will try to share the objects
            (display lists, texture objects, shaders) between the old and new
            contexts.  If this is possible, only the current state of the GL
            context is lost, and the application should simply restore state.

            :event:
            """

        def on_deactivate(self) -> EVENT_HANDLE_STATE:
            """The window was deactivated.

            This event can be triggered by clicking on another application
            window.  When a window is deactivated it no longer has the
            keyboard focus.

            :event:
            """

        def on_draw(self) -> EVENT_HANDLE_STATE:
            """The window contents should be redrawn.

            The `EventLoop` will dispatch this event when the `draw`
            method has been called. The window will already have the
            GL context, so there is no need to call `switch_to`. The window's
            `flip` method will be called immediately after this event,
            so your event handler should not.

            You should make no assumptions about the window contents when
            this event is triggered; a resize or expose event may have
            invalidated the framebuffer since the last time it was drawn.

            :event:
            """

        def on_expose(self) -> EVENT_HANDLE_STATE:
            """A portion of the window needs to be redrawn.

            This event is triggered when the window first appears, and any time
            the contents of the window is invalidated due to another window
            obscuring it.

            There is no way to determine which portion of the window needs
            redrawing.  Note that the use of this method is becoming
            increasingly uncommon, as newer window managers composite windows
            automatically and keep a backing store of the window contents.

            :event:
            """

        def on_file_drop(self, x: int, y: int, paths: list[str]) -> EVENT_HANDLE_STATE:
            """File(s) were dropped into the window.

            Args:
                x:
                    Distance in pixels from the left edge of the window where a drop occurred.
                y:
                    Distance in pixels from the bottom edge of the window where a drop occurred.
                paths:
                    File path strings that were dropped into the window.

            .. versionadded:: 1.5.1

            :event:
            """

        def on_hide(self) -> EVENT_HANDLE_STATE:
            """The window was hidden.

            This event is triggered when a window is minimised
            or hidden by the user.

            :event:
            """

        def on_key_press(self, symbol: int, modifiers: int) -> EVENT_HANDLE_STATE:
            """A key on the keyboard was pressed (and held down).

            Since pyglet 1.1 the default handler dispatches the :py:meth:`~pyglet.window.Window.on_close`
            event if the ``ESC`` key is pressed.

            Args:
                symbol:
                    The key symbol pressed.
                modifiers:
                    Bitwise combination of the key modifiers active.

            :event:
            """

        def on_key_release(self, symbol: int, modifiers: int) -> EVENT_HANDLE_STATE:
            """A key on the keyboard was released.

            Args:
                symbol:
                    The key symbol pressed.
                modifiers:
                    Bitwise combination of the key modifiers active.

            :event:
            """

        def on_mouse_motion(self, x: int, y: int, dx: int, dy: int) -> EVENT_HANDLE_STATE:
            """The mouse was moved with no buttons held down.

            Args:
                x:
                    Distance in pixels from the left edge of the window.
                y:
                    Distance in pixels from the bottom edge of the window.
                dx:
                    Relative X position from the previous mouse position.
                dy: int
                    Relative Y position from the previous mouse position.

            :event:
            """

        def on_mouse_drag(self, x: int, y: int, dx: int, dy: int, buttons: int, modifiers: int) -> EVENT_HANDLE_STATE:
            """The mouse was moved with one or more mouse buttons pressed.

            This event will continue to be fired even if the mouse leaves
            the window, so long as the drag buttons are continuously held down.

            Args:
                x:
                    Distance in pixels from the left edge of the window.
                y:
                    Distance in pixels from the bottom edge of the window.
                dx:
                    Relative X position from the previous mouse position.
                dy:
                    Relative Y position from the previous mouse position.
                buttons:
                    Bitwise combination of the mouse buttons currently pressed.
                modifiers:
                    Bitwise combination of any keyboard modifiers currently active.

            :event:
            """

        def on_mouse_press(self, x: int, y: int, button: int, modifiers: int) -> EVENT_HANDLE_STATE:
            """A mouse button was pressed (and held down).

            Args:
                x:
                    Distance in pixels from the left edge of the window.
                y:
                    Distance in pixels from the bottom edge of the window.
                button:
                    The mouse button that was pressed.
                modifiers:
                    Bitwise combination of any keyboard modifiers currently active.

            :event:
            """

        def on_mouse_release(self, x: int, y: int, button: int, modifiers: int) -> EVENT_HANDLE_STATE:
            """A mouse button was released.

            Args:
                x:
                    Distance in pixels from the left edge of the window.
                y:
                    Distance in pixels from the bottom edge of the window.
                button:
                    The mouse button that was released.
                modifiers:
                    Bitwise combination of any keyboard modifiers currently active.

            :event:
            """

        def on_mouse_scroll(self, x: int, y: int, scroll_x: float, scroll_y: float) -> EVENT_HANDLE_STATE:
            """The mouse wheel was scrolled.

            Note that most mice have only a vertical scroll wheel, so
            scroll_x is usually 0.  An exception to this is the Apple Mighty
            Mouse, which has a mouse ball in place of the wheel which allows
            both ``scroll_x`` and ``scroll_y`` movement.

            Args:
                x:
                    Distance in pixels from the left edge of the window.
                y:
                    Distance in pixels from the bottom edge of the window.
                scroll_x:
                    Amount of movement on the horizontal axis.
                scroll_y:
                    Amount of movement on the vertical axis.

            :event:
            """

        def on_mouse_enter(self, x: int, y: int) -> EVENT_HANDLE_STATE:
            """The mouse was moved into the window.

            This event will not be triggered if the mouse is currently being
            dragged.

            :event:
            """

        def on_mouse_leave(self, x: int, y: int) -> EVENT_HANDLE_STATE:
            """The mouse was moved outside the window.

            This event will not be triggered if the mouse is currently being dragged.

            .. note: Coordinates of the mouse pointer will be outside the window rectangle.

            :event:
            """

        def on_move(self, x: int, y: int) -> EVENT_HANDLE_STATE:
            """The window was moved.

            Args:
                x:
                    Distance from the left edge of the screen to the left edge of the window.
                y:
                    Distance from the top edge of the screen to the *top* edge of the window.

            .. note: This is one of few methods in pyglet which use a Y-down coordinate system.

            :event:
            """

        def on_refresh(self, dt: float) -> EVENT_HANDLE_STATE:
            """The window contents should be redrawn.

            The ``EventLoop`` will dispatch this event when the ``draw``
            method has been called. The window will already have the
            GL context, so there is no need to call :py:meth:`~pyglet.window.Window.switch_to`. The window's
            :py:meth:`~pyglet.window.Window.flip` method will be called immediately after this event, so your
            event handler should not.

            You should make no assumptions about the window contents when
            this event is triggered; a resize or expose event may have
            invalidated the framebuffer since the last time it was drawn.

            .. versionadded:: 2.0

            :event:
            """

        def on_resize(self, width: int, height: int) -> EVENT_HANDLE_STATE:
            """The window was resized.

            The window will have the GL context when this event is dispatched;
            there is no need to call :py:meth:`~pyglet.window.Window.switch_to` in this handler.

            :event:
            """

        def on_scale(self, scale: float, dpi: int) -> EVENT_HANDLE_STATE:
            """A default scale event handler.

            This default handler is called if the screen or system's DPI changes
            during runtime.
            """

        def on_show(self) -> EVENT_HANDLE_STATE:
            """The window was shown.

            This event is triggered when a window is restored after being
            minimised, hidden, or after being displayed for the first time.

            :event:
            """

        def on_text(self, text: str) -> EVENT_HANDLE_STATE:
            """The user input some text.

            Typically, this is called after :py:meth:`~pyglet.window.Window.on_key_press` and before
            :py:meth:`~pyglet.window.Window.on_key_release`, but may also be called multiple times if the key
            is held down (key repeating); or called without key presses if
            another input method was used (e.g., a pen input).

            You should always use this method for interpreting text, as the
            key symbols often have complex mappings to their unicode
            representation which this event takes care of.

            :event:
            """

        def on_text_motion(self, motion: int) -> EVENT_HANDLE_STATE:
            """The user moved the text input cursor.

            Typically this is called after :py:meth:`~pyglet.window.Window.on_key_press` and before
            :py:meth:`~pyglet.window.Window.on_key_release`, but may also be called multiple times if the key
            is help down (key repeating).

            You should always use this method for moving the text input cursor
            (caret), as different platforms have different default keyboard
            mappings, and key repeats are handled correctly.

            The values that `motion` can take are defined in
            :py:mod:`pyglet.window.key`:

            * MOTION_UP
            * MOTION_RIGHT
            * MOTION_DOWN
            * MOTION_LEFT
            * MOTION_NEXT_WORD
            * MOTION_PREVIOUS_WORD
            * MOTION_BEGINNING_OF_LINE
            * MOTION_END_OF_LINE
            * MOTION_NEXT_PAGE
            * MOTION_PREVIOUS_PAGE
            * MOTION_BEGINNING_OF_FILE
            * MOTION_END_OF_FILE
            * MOTION_BACKSPACE
            * MOTION_DELETE
            * MOTION_COPY
            * MOTION_PASTE

            :event:
            """

        def on_text_motion_select(self, motion: int) -> EVENT_HANDLE_STATE:
            """The user moved the text input cursor while extending the selection.

            Typically this is called after :py:meth:`~pyglet.window.Window.on_key_press` and before
            :py:meth:`~pyglet.window.Window.on_key_release`, but may also be called multiple times if the key
            is help down (key repeating).

            You should always use this method for responding to text selection
            events rather than the raw :py:meth:`~pyglet.window.Window.on_key_press`, as different platforms
            have different default keyboard mappings, and key repeats are
            handled correctly.

            The values that `motion` can take are defined in :py:mod:`pyglet.window.key`:

            * MOTION_UP
            * MOTION_RIGHT
            * MOTION_DOWN
            * MOTION_LEFT
            * MOTION_NEXT_WORD
            * MOTION_PREVIOUS_WORD
            * MOTION_BEGINNING_OF_LINE
            * MOTION_END_OF_LINE
            * MOTION_NEXT_PAGE
            * MOTION_PREVIOUS_PAGE
            * MOTION_BEGINNING_OF_FILE
            * MOTION_END_OF_FILE
            * MOTION_COPY
            * MOTION_PASTE

            :event:
            """


BaseWindow.register_event_type('on_key_press')
BaseWindow.register_event_type('on_key_release')
BaseWindow.register_event_type('on_text')
BaseWindow.register_event_type('on_text_motion')
BaseWindow.register_event_type('on_text_motion_select')
BaseWindow.register_event_type('on_mouse_motion')
BaseWindow.register_event_type('on_mouse_drag')
BaseWindow.register_event_type('on_mouse_press')
BaseWindow.register_event_type('on_mouse_release')
BaseWindow.register_event_type('on_mouse_scroll')
BaseWindow.register_event_type('on_mouse_enter')
BaseWindow.register_event_type('on_mouse_leave')
BaseWindow.register_event_type('on_close')
BaseWindow.register_event_type('on_expose')
BaseWindow.register_event_type('_on_internal_resize')
BaseWindow.register_event_type('_on_internal_scale')
BaseWindow.register_event_type('on_resize')
BaseWindow.register_event_type('on_scale')
BaseWindow.register_event_type('on_move')
BaseWindow.register_event_type('on_activate')
BaseWindow.register_event_type('on_deactivate')
BaseWindow.register_event_type('on_show')
BaseWindow.register_event_type('on_hide')
BaseWindow.register_event_type('on_context_lost')
BaseWindow.register_event_type('on_context_state_lost')
BaseWindow.register_event_type('on_file_drop')
BaseWindow.register_event_type('on_draw')
BaseWindow.register_event_type('on_refresh')


class FPSDisplay:
    """Display of a window's framerate.

    This is a convenience class to aid in profiling and debugging.  Typical
    usage is to create an `FPSDisplay` for each window, and draw the display
    at the end of the windows' :py:meth:`~pyglet.window.Window.on_draw` event handler::

        from pyglet.window import Window, FPSDisplay

        window = Window()
        fps_display = FPSDisplay(window)

        @window.event
        def on_draw():
            # ... perform ordinary window drawing operations ...

            fps_display.draw()

    The style and position of the display can be modified via the ``label`` attribute. The display can be set to
    update more or less often by setting the `update_period` attribute.

    .. note: Setting the `update_period` to a value smaller than your Window refresh rate will cause
             inaccurate readings.
    """

    #: Time in seconds between updates.
    update_period = 0.25

    #: The text label displaying the framerate.
    label: Label

    def __init__(self, window: pyglet.window.Window, color: tuple[int, int, int, int] = (127, 127, 127, 127),
                 samples: int = 240) -> None:
        """Create an FPS Display.

        Args:
            window:
                The Window you wish to display frame rate for.
            color:
                The RGBA color of the text display. Each channel represented as 0-255.
            samples:
                How many delta samples are used to calculate the mean FPS.
        """
        from collections import deque
        from statistics import mean
        from time import time

        from pyglet.text import Label
        self._time = time
        self._mean = mean

        # Hook into the Window.flip method:
        self._window_flip, window.flip = window.flip, self._hook_flip
        self.label = Label('', x=10, y=10, font_size=24, weight='bold', color=color)

        self._elapsed = 0.0
        self._last_time = time()
        self._delta_times = deque(maxlen=samples)

    def update(self) -> None:
        """Records a new data point at the current time.

        This method is called automatically when the window buffer is flipped.
        """
        t = self._time()
        delta = t - self._last_time
        self._elapsed += delta
        self._delta_times.append(delta)
        self._last_time = t

        if self._elapsed >= self.update_period:
            self._elapsed = 0
            self.label.text = f'{1 / self._mean(self._delta_times):.2f}'

    def draw(self) -> None:
        """Draw the label."""
        self.label.draw()

    def _hook_flip(self) -> None:
        self.update()
        self._window_flip()


if _is_pyglet_doc_run:
    # We are building documentation. Trick docs into thinking BaseWindow is Window.
    import inspect

    Window = BaseWindow
    Window.__name__ = 'Window'
    Window.__qualname__ = 'Window'

    # We also need to replace all qualname members so Sphinx and Typing modules pick up the correct class.
    for _, method_obj in inspect.getmembers(Window, predicate=inspect.isfunction):
        method_obj.__qualname__ = method_obj.__qualname__.replace('BaseWindow', 'Window')

else:
    # Try to determine which platform to use.
    if pyglet.options['headless']:
        from pyglet.window.headless import HeadlessWindow as Window
    elif pyglet.compat_platform == 'darwin':
        from pyglet.window.cocoa import CocoaWindow as Window
    elif pyglet.compat_platform in ('win32', 'cygwin'):
        from pyglet.window.win32 import Win32Window as Window
    else:
        from pyglet.window.xlib import XlibWindow as Window

# Create shadow window. (trickery is for circular import)
if not _is_pyglet_doc_run:
    pyglet.window = sys.modules[__name__]
    gl._create_shadow_window()  # noqa: SLF001


__all__ = (
    # imported
    "event",
    "key",
    # classes
    "BaseWindow",
    "Window",
    "MouseCursor",
    "DefaultMouseCursor",
    "ImageMouseCursor",
    "FPSDisplay",
    # errors
    "WindowException",
    "NoSuchScreenModeException",
    "NoSuchDisplayException",
    "NoSuchConfigException",
    "MouseCursorException",

)
